from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas import PRECISION

from compas.geometry import centroid_points
from compas.geometry import normal_polygon

from compas.geometry import distance_point_point
from compas.geometry import distance_point_line
from compas.geometry import distance_point_plane
from compas.geometry import is_point_on_line
from compas.geometry import is_point_on_segment
from compas.geometry import is_point_on_polyline
from compas.geometry import is_point_in_triangle
from compas.geometry import is_point_in_circle
from compas.geometry import is_polygon_convex_xy
from compas.geometry import is_point_in_polygon_xy
from compas.geometry import is_point_in_convex_polygon_xy
from compas.geometry import is_point_behind_plane

from compas.geometry import transform_points

from compas.geometry.primitives import Primitive
from compas.geometry.primitives import Vector


class Point(Primitive):
    """A point is defined by XYZ coordinates.

    Parameters
    ----------
    x : float
        The X coordinate of the point.
    y : float
        The Y coordinate of the point.
    z : float, optional
        The Z coordinate of the point.

    Attributes
    ----------
    x : float
        The X coordinate of the point.
    y : float
        The Y coordinate of the point.
    z : float
        The Z coordinate of the point.

    Notes
    -----
    A `Point` object supports direct access to its xyz coordinates through
    the dot notation, as well list-style access using indices. Indexed
    access is implemented such that the `Point` behaves like a circular
    list [1]_.

    References
    ----------
    .. [1] Stack Overflow. *Pythonic Circular List*.
           Available at: https://stackoverflow.com/questions/8951020/pythonic-circular-list.

    Examples
    --------
    >>> p1 = Point(1, 2, 3)
    >>> p2 = Point(4, 5, 6)

    The XYZ coordinates of point objects can be accessed as object attributes or by trating the points as lists.

    >>> p1.x
    1.0
    >>> p1.y
    2.0
    >>> p1.z
    3.0
    >>> p1[0]
    1.0
    >>> p1[1]
    2.0
    >>> p1[2]
    3.0

    Point objects support basic arithmetic operations.

    >>> p1 + p2
    Point(5.000, 7.000, 9.000)
    >>> p1 * 2
    Point(2.000, 4.000, 6.000)
    >>> p1 ** 2
    Point(1.000, 4.000, 9.000)
    >>> p1
    Point(1.000, 2.000, 3.000)

    Points and lists can be used interchangeably.

    >>> p1 + [4, 5, 6]
    Point(5.000, 7.000, 9.000)

    Arithmetic operations can also be applied to modify a point object in-place.

    >>> p1 += p2
    >>> p1 *= 2
    >>> p1 **= 2
    >>> p1
    Point(100.000, 196.000, 324.000)

    """

    __slots__ = ['_x', '_y', '_z']

    def __init__(self, x, y, z=0.0, **kwargs):
        super(Point, self).__init__(**kwargs)
        self._x = 0.0
        self._y = 0.0
        self._z = 0.0
        self.x = x
        self.y = y
        self.z = z

    # ==========================================================================
    # data
    # ==========================================================================

    @property
    def DATASCHEMA(self):
        """:class:`schema.Schema` : Schema of the data representation."""
        from schema import Schema
        from compas.data import is_float3
        return Schema(is_float3)

    @property
    def JSONSCHEMANAME(self):
        """str : Name of the schema of the data representation in JSON format."""
        return 'point'

    @property
    def data(self):
        """dict : The data representing the point."""
        return list(self)

    @data.setter
    def data(self, data):
        self.x = data[0]
        self.y = data[1]
        self.z = data[2]

    @classmethod
    def from_data(cls, data):
        """Construct a point from a data dict.

        Parameters
        ----------
        data : dict
            The data dictionary.

        Returns
        -------
        :class:`compas.geometry.Point`
            The constructed point.

        Examples
        --------
        >>> point = Point.from_data([0.0, 0.0, 0.0])
        >>> point
        Point(0.000, 0.000, 0.000)

        """
        return cls(*data)

    # ==========================================================================
    # properties
    # ==========================================================================

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        self._x = float(x)

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, y):
        self._y = float(y)

    @property
    def z(self):
        return self._z

    @z.setter
    def z(self, z):
        self._z = float(z)

    # ==========================================================================
    # customization
    # ==========================================================================

    def __repr__(self):
        return 'Point({0:.{3}f}, {1:.{3}f}, {2:.{3}f})'.format(self.x, self.y, self.z, PRECISION[:1])

    def __len__(self):
        return 3

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [self[i] for i in range(*key.indices(len(self)))]
        i = key % 3
        if i == 0:
            return self.x
        if i == 1:
            return self.y
        if i == 2:
            return self.z
        raise KeyError

    def __setitem__(self, key, value):
        i = key % 3
        if i == 0:
            self.x = value
            return
        if i == 1:
            self.y = value
            return
        if i == 2:
            self.z = value
            return
        raise KeyError

    def __iter__(self):
        return iter([self.x, self.y, self.z])

    def __eq__(self, other):
        """Is this point equal to the other point?

        Two points are considered equal if their XYZ coordinates are identical.

        Parameters
        ----------
        other : [float, float, float] | :class:`compas.geometry.Point`
            The point to compare.

        Returns
        -------
        bool
            True if the points are equal.
            False otherwise.

        """
        return self.x == other[0] and self.y == other[1] and self.z == other[2]

    def __add__(self, other):
        """Return a point that is the sum of this point and another point.

        Parameters
        ----------
        other : [float, float, float] | :class:`compas.geometry.Point`
            The point to add.

        Returns
        -------
        :class:`compas.geometry.Point`
            The resulting new point.

        """
        return Point(self.x + other[0], self.y + other[1], self.z + other[2])

    def __sub__(self, other):
        """Return a vector` that is the the difference between this point
        and another point.

        Parameters
        ----------
        other : [float, float, float] | :class:`compas.geometry.Point`
            The point to subtract.

        Returns
        -------
        :class:`compas.geometry.Vector`
            A vector from other to self.

        """
        x = self.x - other[0]
        y = self.y - other[1]
        z = self.z - other[2]
        return Vector(x, y, z)

    def __mul__(self, n):
        """Create a point from the coordinates of the current point multiplied
        by the given factor.

        Parameters
        ----------
        n : float
            The multiplication factor.

        Returns
        -------
        :class:`compas.geometry.Point`
            The resulting new point.

        """
        return Point(n * self.x, n * self.y, n * self.z)

    def __truediv__(self, n):
        """Create a point from the coordinates of the current point
        divided by the given factor.

        Parameters
        ----------
        n : float
            The multiplication factor.

        Returns
        -------
        :class:`compas.geometry.Point`
            The resulting new point.

        """
        return Point(self.x / n, self.y / n, self.z / n)

    def __pow__(self, n):
        """Create a point from the coordinates of the current point raised
        to the given power.

        Parameters
        ----------
        n : float
            The power.

        Returns
        -------
        :class:`compas.geometry.Point`
            A new point with raised coordinates.

        """
        return Point(self.x ** n, self.y ** n, self.z ** n)

    def __iadd__(self, other):
        """Add the coordinates of the other point to this point.

        Parameters
        ----------
        other : [float, float, float] | :class:`compas.geometry.Point`
            The point to add.

        Returns
        -------
        None

        """
        self.x += other[0]
        self.y += other[1]
        self.z += other[2]
        return self

    def __isub__(self, other):
        """Subtract the coordinates of the other point from this point.

        Parameters
        ----------
        other : [float, float, float] | :class:`compas.geometry.Point`
            The point to subtract.

        Returns
        -------
        None

        """
        self.x -= other[0]
        self.y -= other[1]
        self.z -= other[2]
        return self

    def __imul__(self, n):
        """Multiply the coordinates of this point by the given factor.

        Parameters
        ----------
        n : float
            The multiplication factor.

        Returns
        -------
        None

        """
        self.x *= n
        self.y *= n
        self.z *= n
        return self

    def __itruediv__(self, n):
        """Divide the coordinates of this point by the given factor.

        Parameters
        ----------
        n : float
            The multiplication factor.

        Returns
        -------
        None

        """
        self.x /= n
        self.y /= n
        self.z /= n
        return self

    def __ipow__(self, n):
        """Raise the coordinates of this point to the given power.

        Parameters
        ----------
        n : float
            The power.

        Returns
        -------
        None

        """
        self.x **= n
        self.y **= n
        self.z **= n
        return self

    # ==========================================================================
    # constructors
    # ==========================================================================

    # ==========================================================================
    # static
    # ==========================================================================

    @staticmethod
    def transform_collection(collection, X):
        """Transform a collection of points.

        Parameters
        ----------
        collection : list[[float, float, float] | :class:`compas.geometry.Point`]
            The collection of points.

        Returns
        -------
        None
            The points are modified in-place.

        Examples
        --------
        >>> from compas.geometry import Translation
        >>> T = Translation.from_vector([1.0, 2.0, 3.0])
        >>> a = Point(0.0, 0.0, 0.0)
        >>> points = [a]
        >>> Point.transform_collection(points, T)
        >>> b = points[0]
        >>> b
        Point(1.000, 2.000, 3.000)
        >>> a is b
        True

        """
        data = transform_points(collection, X)
        for point, xyz in zip(collection, data):
            point.x = xyz[0]
            point.y = xyz[1]
            point.z = xyz[2]

    @staticmethod
    def transformed_collection(collection, X):
        """Create a collection of transformed points.

        Parameters
        ----------
        collection : list[[float, float, float] | :class:`compas.geometry.Point`]
            The collection of points.

        Returns
        -------
        list[:class:`compas.geometry.Point`]
            The transformed points.

        Examples
        --------
        >>> from compas.geometry import Translation
        >>> T = Translation.from_vector([1.0, 2.0, 3.0])
        >>> a = Point(0.0, 0.0, 0.0)
        >>> points = [a]
        >>> points = Point.transformed_collection(points, T)
        >>> b = points[0]
        >>> b
        Point(1.000, 2.000, 3.000)
        >>> a is b
        False

        """
        points = [point.copy() for point in collection]
        Point.transform_collection(points, X)
        return points

    # ==========================================================================
    # methods
    # ==========================================================================

    def distance_to_point(self, point):
        """Compute the distance to another point.

        Parameters
        ----------
        point : [float, float, float] | :class:`compas.geometry.Point`
            The other point.

        Returns
        -------
        float
            The distance.

        Examples
        --------
        >>> a = Point(0.0, 0.0, 0.0)
        >>> b = Point(1.0, 0.0, 0.0)
        >>> a.distance_to_point(b)
        1.0

        """
        return distance_point_point(self, point)

    def distance_to_line(self, line):
        """Compute the distance to a line.

        Parameters
        ----------
        line : [point, point] | :class:`compas.geometry.Line`
            The line.

        Returns
        -------
        float
            The distance.

        Examples
        --------
        >>> from compas.geometry import Line
        >>> point = Point(0.0, 0.0, 0.0)
        >>> line = Line(Point(1.0, 0.0, 0.0), Point(1.0, 1.0, 0.0))
        >>> point.distance_to_line(line)
        1.0

        """
        return distance_point_line(self, line)

    def distance_to_plane(self, plane):
        """Compute the distance to a plane.

        Parameters
        ----------
        plane : [point, vector] | :class:`compas.geometry.Plane`
            The plane.

        Returns
        -------
        float
            The distance.

        Examples
        --------
        >>> from compas.geometry import Plane
        >>> from compas.geometry import Vector
        >>> point = Point(0.0, 0.0, 0.0)
        >>> plane = Plane(Point(1.0, 0.0, 0.0), Vector(1.0, 0.0, 0.0))
        >>> point.distance_to_plane(plane)
        1.0

        """
        return distance_point_plane(self, plane)

    # ==========================================================================
    # 2D predicates
    # ==========================================================================

    def in_polygon(self, polygon, convex=None):
        """Determine if the point lies inside the given polygon.

        Parameters
        ----------
        polygon : sequence[point] | :class:`compas.geometry.Polygon`
            The polygon.
        convex : Literal[None, True, False], optional
            If None, determine if the polygon is convex.
            If False, use the non-convex algorithm.
            If True, use the convex algorithm.

        Returns
        -------
        bool
            True, if the point lies in the polygon.
            False, otherwise.

        Examples
        --------
        >>> from compas.geometry import Polygon
        >>> poly = Polygon([Point(0.0, 0.0, 0.0), Point(1.0, 0.0, 0.0), Point(1.0, 1.0, 0.0), Point(0.0, 1.0, 0.0)])
        >>> point = Point(0.5, 0.5, 0.0)
        >>> point.in_polygon(poly)
        True

        """
        if convex is None:
            convex = is_polygon_convex_xy(polygon)
        if convex:
            return is_point_in_convex_polygon_xy(self, polygon)
        return is_point_in_polygon_xy(self, polygon)

    # ==========================================================================
    # 3D predicates
    # ==========================================================================

    def on_line(self, line):
        """Determine if the point lies on the given line.

        Parameters
        ----------
        line : [point, point] | :class:`compas.geometry.Line`
            The line.

        Returns
        -------
        bool
            True, if the point lies on the line.
            False, otherwise.

        Examples
        --------
        >>> from compas.geometry import Line
        >>> line = Line(Point(1.0, 0.0, 0.0), Point(1.0, 1.0, 0.0))
        >>> point = line.point(1.5)
        >>> point.on_line(line)
        True

        """
        return is_point_on_line(self, line)

    def on_segment(self, segment):
        """Determine if the point lies on the given segment.

        Parameters
        ----------
        segment : [point, point] | :class:`compas.geometry.Line`
            The segment.

        Returns
        -------
        bool
            True, if the point lies on the segment.
            False, otherwise.

        Examples
        --------
        >>> from compas.geometry import Line
        >>> line = Line(Point(1.0, 0.0, 0.0), Point(1.0, 1.0, 0.0))
        >>> point = line.point(1.5)
        >>> point.on_segment(line)
        False

        """
        return is_point_on_segment(self, segment)

    def on_polyline(self, polyline):
        """Determine if the point lies on the given polyline.

        Parameters
        ----------
        polyline : sequence[point] | :class:`compas.geometry.Polyline`
            The polyline.

        Returns
        -------
        bool
            True, if the point lies on the polyline.
            False, otherwise.

        Examples
        --------
        >>> from compas.geometry import Polyline
        >>> poly = Polyline([Point(0.0, 0.0, 0.0), Point(1.0, 0.0, 0.0), Point(2.0, 0.0, 0.0)])
        >>> point = poly.point(0.5)
        >>> point.on_polyline(poly)
        True

        """
        return is_point_on_polyline(self, polyline)

    def on_circle(self, circle):
        """Determine if the point lies on the given circle.

        Parameters
        ----------
        circle : [plane, radius] | :class:`compas.geometry.Circle`
            The circle.

        Returns
        -------
        bool
            True, if the point lies on the circle.
            False, otherwise.

        """
        raise NotImplementedError

    def in_triangle(self, triangle):
        """Determine if the point lies inside the given triangle.

        Parameters
        ----------
        triangle : [point, point, point] | :class:`compas.geometry.Polygon`
            The triangle.

        Returns
        -------
        bool
            True, if the point lies in the triangle.
            False, otherwise.

        Examples
        --------
        >>> from compas.geometry import Polygon
        >>> tri = Polygon([Point(0.0, 0.0, 0.0), Point(1.0, 0.0, 0.0), Point(0.5, 1.0, 0.0)])
        >>> point = Point(0.5, 0.5, 0.0)
        >>> point.in_triangle(tri)
        True

        """
        return is_point_in_triangle(self, triangle)

    def in_circle(self, circle):
        """Determine if the point lies inside the given circle.

        Parameters
        ----------
        circle : [plane, radius] | :class:`compas.geometry.Circle`
            The circle.

        Returns
        -------
        bool
            True, if the point lies inside the circle.
            False, otherwise.

        Examples
        --------
        >>> from compas.geometry import Vector
        >>> from compas.geometry import Plane
        >>> from compas.geometry import Circle
        >>> circle = Circle(Plane(Point(0.0, 0.0, 0.0), Vector(0.0, 0.0, 1.0)), 1.0)
        >>> point = Point(0.5, 0.0, 0.0)
        >>> point.in_circle(circle)
        True

        """
        return is_point_in_circle(self, circle)

    def in_polyhedron(self, polyhedron):
        """Determine if the point lies inside the given polyhedron.

        Parameters
        ----------
        polyhedron : [vertices, faces] | :class:`compas.geometry.Polyhedron`
            The polyhedron.

        Returns
        -------
        bool
            True, if the point lies on the polyline.
            False, otherwise.

        """
        vertices, faces = polyhedron
        polygons = [[vertices[index] for index in face] for face in faces]
        planes = [[centroid_points(polygon), normal_polygon(polygon)] for polygon in polygons]
        return all(is_point_behind_plane(self, plane) for plane in planes)

    # ==========================================================================
    # transformations
    # ==========================================================================

    def transform(self, T):
        """Transform this point.

        Parameters
        ----------
        T : :class:`compas.geometry.Transformation` | list[list[float]]
            The transformation matrix.

        Examples
        --------
        >>> from compas.geometry import Translation
        >>> point = Point(0.0, 0.0, 0.0)
        >>> T = Translation.from_vector([1.0, 1.0, 1.0])
        >>> point.transform(T)
        >>> point.x == 1.0
        True

        """
        point = transform_points([self], T)[0]
        self.x = point[0]
        self.y = point[1]
        self.z = point[2]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import Rhino

from compas.geometry import Line
from compas.geometry import Circle
from compas.geometry import Ellipse

from ._exceptions import ConversionError

from ._curves import circle_to_rhino_curve
from ._curves import ellipse_to_rhino_curve
from ._curves import line_to_rhino_curve
from ._curves import curve_to_compas_circle
from ._curves import curve_to_compas_ellipse
from ._curves import curve_to_compas_polyline
from ._curves import curve_to_compas_line

from ._geometry import RhinoGeometry


class RhinoCurve(RhinoGeometry):
    """Wrapper for Rhino curves."""

    @property
    def geometry(self):
        return self._geometry

    @geometry.setter
    def geometry(self, geometry):
        """Set the geometry of the wrapper.

        Parameters
        ----------
        geometry : :rhino:`Rhino_Geometry_Curve` or :class:`compas.geometry.Line` or :class:`compas.geometry.Circle` or :class:`compas.geometry.Ellipse`
            The geometry object defining a curve.

        Raises
        ------
        :class:`ConversionError`
            If the geometry cannot be converted to a curve.
        """
        if not isinstance(geometry, Rhino.Geometry.Curve):
            if isinstance(geometry, Line):
                geometry = line_to_rhino_curve(geometry)
            elif isinstance(geometry, Circle):
                geometry = circle_to_rhino_curve(geometry)
            elif isinstance(geometry, Ellipse):
                geometry = ellipse_to_rhino_curve(geometry)
            else:
                raise ConversionError('The geometry cannot be converted to a curve.')
        self._geometry = geometry

    def to_compas(self):
        """Convert the curve to a COMPAS curve.

        Returns
        -------
        :class:`compas_rhino.geometry.RhinoNurbsCurve`
        """
        from compas.geometry import NurbsCurve
        curve = NurbsCurve()
        curve.rhino_curve = self.geometry
        return curve

    def to_compas_circle(self):
        """Convert the curve to a COMPAS circle.

        Returns
        -------
        :class:`compas.geometry.Circle`
        """
        return curve_to_compas_circle(self.geometry)

    def to_compas_ellipse(self):
        """Convert the curve to a COMPAS ellipse.

        Returns
        -------
        :class:`compas.geometry.Ellipse`
        """
        return curve_to_compas_ellipse(self.geometry)

    def to_compas_line(self):
        """Convert the curve to a COMPAS line.

        Returns
        -------
        :class:`compas.geometry.Line`
        """
        return curve_to_compas_line(self.geometry)

    def to_compas_polyline(self):
        """Convert the curve to a COMPAS polyline.

        Returns
        -------
        :class:`compas.geometry.Polyline`
        """
        return curve_to_compas_polyline(self.geometry)

    def closest_point(self, point, maxdist=0.0, return_param=False):
        """Compute the closest point on a curve to a point in space.

        Parameters
        ----------
        point : point
            A point location.
        maxdist : float, optional
            The maximum distance between the point on the curve and the curve.
            Default is ``0.0``.
        return_param : bool, optional
            Return not only the point coordinates, but also the parameter of the point on the curve.
            Default is False.

        Returns
        -------
        list
            The XYZ coordinates of the closest point, if ``return_param`` is False.
            The XYZ coordinates of the closest point and the curve parameter, if ``return_param`` is True.

        """
        rc, t = self.geometry.ClosestPoint(Rhino.Geometry.Point3d(*point), maxdist)
        x, y, z = list(self.geometry.PointAt(t))
        if not return_param:
            return [x, y, z]
        return [x, y, z, t]

    def closest_points(self, points, maxdist=0.0):
        """Compute the closest points on the curve to a list of point locations.

        Parameters
        ----------
        points : list
            The point locations.
        maxdist : float, optional
            The maximum distance between the closest points and the curve.
            Default is ``0.0``.

        Returns
        -------
        list
            A list of closest point locations.
        """
        return [self.closest_point(point, maxdist) for point in points]



"""
********************************************************************************
base
********************************************************************************

.. deprecated:: 1.5
    Use `compas.data` instead

.. currentmodule:: compas.base

Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Base

"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import warnings

from compas.data import Data

Base = Data

__all__ = [
    'Base',
]

warnings.warn(
    "The base module is deprecated. Use the data module instead",
    DeprecationWarning,
    stacklevel=2
)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas_ghpython
from compas.artists import PrimitiveArtist
from .artist import GHArtist


class LineArtist(GHArtist, PrimitiveArtist):
    """Artist for drawing lines.

    Parameters
    ----------
    line : :class:`compas.geometry.Line`
        A COMPAS line.
    **kwargs : dict, optional
        Additional keyword arguments.
        See :class:`compas_ghpython.artists.GHArtist` and :class:`compas.artists.PrimitiveArtist` for more info.

    """

    def __init__(self, line, **kwargs):
        super(LineArtist, self).__init__(primitive=line, **kwargs)

    def draw(self):
        """Draw the line.

        Returns
        -------
        :rhino:`Rhino.Geometry.Line`

        """
        lines = [self._get_args(self.primitive)]
        return compas_ghpython.draw_lines(lines)[0]

    @staticmethod
    def _get_args(primitive):
        start = list(primitive.start)
        end = list(primitive.end)
        return {'start': start, 'end': end}



"""
************
compas_rhino
************

.. currentmodule:: compas_rhino

.. toctree::
    :maxdepth: 1

    compas_rhino.artists
    compas_rhino.conduits
    compas_rhino.conversions
    compas_rhino.forms
    compas_rhino.geometry
    compas_rhino.utilities

"""
from __future__ import absolute_import

import io
import os

import compas
import compas._os

if compas.is_rhino():
    import rhinoscriptsyntax as rs  # noqa: F401
    from .utilities import *  # noqa: F401 F403


__version__ = '1.14.0'


PURGE_ON_DELETE = True

INSTALLABLE_PACKAGES = ['compas', 'compas_rhino', 'compas_ghpython']
SUPPORTED_VERSIONS = ['5.0', '6.0', '7.0', '8.0']
DEFAULT_VERSION = '7.0'

IRONPYTHON_PLUGIN_GUID = '814d908a-e25c-493d-97e9-ee3861957f49'
GRASSHOPPER_PLUGIN_GUID = 'b45a29b1-4343-4035-989e-044e8580d9cf'
RHINOCYCLES_PLUGIN_GUID = '9bc28e9e-7a6c-4b8f-a0c6-3d05e02d1b97'


__all__ = [
    'PURGE_ON_DELETE',
    'INSTALLABLE_PACKAGES',
    'SUPPORTED_VERSIONS',
    'DEFAULT_VERSION',
    'IRONPYTHON_PLUGIN_GUID',
    'GRASSHOPPER_PLUGIN_GUID',
    'RHINOCYCLES_PLUGIN_GUID',

    'clear',
    'redraw'
]

__all_plugins__ = [
    'compas_rhino.geometry.booleans',
    'compas_rhino.geometry.trimesh',
    'compas_rhino.install',
    'compas_rhino.uninstall',
    'compas_rhino.artists',
    'compas_rhino.geometry.curves',
    'compas_rhino.geometry.surfaces',
]


# =============================================================================
# General helpers
# =============================================================================


def clear():
    guids = get_objects()  # noqa: F405
    delete_objects(guids, purge=True)  # noqa: F405


def redraw():
    rs.EnableRedraw(True)
    rs.Redraw()


def _check_rhino_version(version):
    if not version:
        return DEFAULT_VERSION

    if version not in SUPPORTED_VERSIONS:
        raise Exception('Unsupported Rhino version: {}'.format(version))

    return version


def _get_package_path(package):
    return os.path.abspath(os.path.dirname(package.__file__))


def _get_bootstrapper_path(install_path):
    return os.path.join(install_path, 'compas_bootstrapper.py')


def _get_bootstrapper_data(compas_bootstrapper):
    data = {}

    if not os.path.exists(compas_bootstrapper):
        return data

    content = io.open(compas_bootstrapper, encoding='utf8').read()
    exec(content, data)

    return data


def _try_remove_bootstrapper(path):
    """Try to remove bootstrapper.

    Returns
    -------
    bool: True if the operation did not cause errors, False otherwise.
    """

    bootstrapper = _get_bootstrapper_path(path)

    if os.path.exists(bootstrapper):
        try:
            os.remove(bootstrapper)
            return True
        except:  # noqa: E722
            return False
    return True


# =============================================================================
# Base Application folder (Program Files on Windows and Applications on Mac)
# =============================================================================


def _get_rhino_application_folder(version):
    version = _check_rhino_version(version)
    version = version.split('.')[0]   # take the major only

    if compas.WINDOWS:
        app = os.path.join(os.getenv('ProgramFiles'), 'Rhino {}'.format(version))
    elif compas.OSX:
        paths = {
            '5': ['/', 'Applications', 'Rhinoceros.app'],
            '6': ['/', 'Applications', 'Rhinoceros.app'],
            '7': ['/', 'Applications', 'Rhino 7.app', ],
            '8': ['/', 'Applications', 'RhinoWIP.app', ]
        }
        app = os.path.join(*paths[version])

    else:
        raise Exception('Unsupported platform')

    if not os.path.exists(app):
        raise Exception("The application folder does not exist in this location: {}".format(app))

    return app


# =============================================================================
# Base AppData folder (APPDATA on Windows and Application Support on Mac)
# =============================================================================


def _get_rhino_appdata_folder():
    if compas.WINDOWS:
        app = os.path.join(os.getenv('APPDATA'), 'McNeel', 'Rhinoceros')

    elif compas.OSX:
        app = os.path.join(os.getenv('HOME'), 'Library', 'Application Support', 'McNeel', 'Rhinoceros')

    else:
        raise Exception('Unsupported platform')

    if not os.path.exists(app):
        raise Exception("The appdata folder does not exist in this location: {}".format(app))

    return app


# =============================================================================
# Scripts folder
# =============================================================================


def _get_rhino_scripts_path(version):
    appdata = _get_rhino_appdata_folder()
    version = _check_rhino_version(version)
    scripts_path = os.path.join(appdata, '{}'.format(version), 'scripts')

    if not os.path.exists(scripts_path):
        raise Exception("The scripts folder does not exist in this location: {}".format(scripts_path))

    return scripts_path


# =============================================================================
# Managed Plugins folder
# =============================================================================


def _get_rhino_managedplugins_path(version):
    app = _get_rhino_application_folder(version)

    if compas.WINDOWS:
        managedplugins_path = os.path.join(app, 'Plug-ins')

    elif compas.OSX:
        managedplugins_path = os.path.join(app, 'Contents', 'Frameworks', 'RhCore.framework', 'Versions', 'A', 'Resources', 'ManagedPlugIns')

    if not os.path.exists(managedplugins_path):
        raise Exception("The Managed Plug-ins folder does not exist in this location: {}".format(managedplugins_path))

    return managedplugins_path


# =============================================================================
# Plugins folder
# =============================================================================


def _get_rhino_plugins_path(version):
    appdata = _get_rhino_appdata_folder()
    version = _check_rhino_version(version)

    if compas.WINDOWS:
        plugins_path = os.path.join(appdata, '{}'.format(version), 'Plug-ins')

    elif compas.OSX:
        if version == '5.0':
            plugins_path = os.path.join(appdata, 'MacPlugIns')
        else:
            plugins_path = os.path.join(appdata, '{}'.format(version), 'Plug-ins')

    if not os.path.exists(plugins_path):
        raise Exception("The plugins folder does not exist in this location: {}".format(plugins_path))

    return plugins_path


# =============================================================================
# PythonPlugins folder
# =============================================================================


def _get_rhino_pythonplugins_path(version):
    version = _check_rhino_version(version)
    return os.path.join(_get_rhino_plugins_path(version), 'PythonPlugins')


# =============================================================================
# IronPython Plugin
# =============================================================================


def _get_rhino_ironpythonplugin_path(version):
    version = _check_rhino_version(version)
    return os.path.join(_get_rhino_plugins_path(version), 'IronPython ({})'.format(IRONPYTHON_PLUGIN_GUID))


# =============================================================================
# Grasshopper
# =============================================================================


def _get_rhino_grasshopperplugin_path(version):
    version = _check_rhino_version(version)

    if compas.WINDOWS:
        gh_path = os.path.join(os.getenv('APPDATA'), 'Grasshopper')

    elif compas.OSX:
        gh_path = os.path.join(_get_rhino_plugins_path(version), 'Grasshopper ({})'.format(GRASSHOPPER_PLUGIN_GUID))

    else:
        raise Exception('Unsupported platform')

    if not os.path.exists(gh_path):
        raise Exception("The grasshopper folder does not exist in this location: {}".format(gh_path))

    return gh_path


# =============================================================================
# IronPython lib folder
# =============================================================================


def _get_rhino_ironpython_lib_path(version):
    version = _check_rhino_version(version)

    if compas.WINDOWS:
        ipy_lib_path = _get_rhino_ironpython_lib_path_win32(version)

    elif compas.OSX:
        ipy_lib_path = _get_rhino_ironpython_lib_path_mac(version)

    if not os.path.exists(ipy_lib_path):
        ipy_lib_path = None
        # print("The lib folder for IronPython does not exist in this location: {}".format(ipy_lib_path))

    return ipy_lib_path


def _get_rhino_ironpython_lib_path_win32(version):
    return os.path.join(_get_rhino_ironpythonplugin_path(version), 'settings', 'lib')


# For 5.0 this is correct
# For +6 we should switch to the same path as on windows
# which is not in the managed plugins but in the appdata plugins
def _get_rhino_ironpython_lib_path_mac(version):
    return os.path.join(_get_rhino_managedplugins_path(version), 'RhinoDLR_Python.rhp', 'Lib')



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import itertools

from compas.files.gltf.constants import VERTEX_COUNT_BY_MODE
from compas.files.gltf.data_classes import PrimitiveData
from compas.files.gltf.helpers import get_weighted_mesh_vertices
from compas.files.gltf.helpers import get_unweighted_primitive_vertices
from compas.files.gltf.helpers import get_mode


class GLTFMesh(object):
    """Object containing mesh data in a format compatible with the glTF standard.

    Attributes
    ----------
    mesh_name : str
        String of the name of the mesh.
    weights : list
        List containing the weights to be applied to morph targets.
    primitive_data_list : list
        List of objects defining the geometry and material of the mesh.
    extras : object
    extensions : object
    context : GLTFContent
        GLTF context in which the mesh exists.
    key : int
        Key of the mesh used in :attr:`compas.files.GLTFMesh.context.meshes`.
    vertices : list
        List of xyz-tuples representing the points of the mesh.
    faces : list
        List of tuples referencing the indices of :attr:`compas.files.GLTFMesh.vertices`
        representing faces of the mesh.

    """
    def __init__(self, primitive_data_list, context, mesh_name=None, weights=None, extras=None, extensions=None):
        self.mesh_name = mesh_name
        self.weights = weights
        self.primitive_data_list = primitive_data_list
        self.extras = extras
        self.extensions = extensions

        self._key = None
        self.context = context
        self._set_key()

    def _set_key(self):
        key = len(self.context.meshes)
        while key in self.context.meshes:
            key += 1
        self.context.meshes[key] = self
        self._key = key

    @property
    def key(self):
        return self._key

    @property
    def vertices(self):
        if not self.weights:
            return get_unweighted_primitive_vertices(self.primitive_data_list)
        return get_weighted_mesh_vertices(self, self.weights)

    @property
    def faces(self):
        faces = []
        shift = 0
        for primitive_data in self.primitive_data_list:
            shifted_indices = self.shift_indices(primitive_data.indices, shift)
            group_size = VERTEX_COUNT_BY_MODE[primitive_data.mode]
            grouped_indices = self.group_indices(shifted_indices, group_size)
            faces.extend(grouped_indices)
            shift += len(primitive_data.attributes['POSITION'])
        return faces

    def shift_indices(self, indices, shift):
        """Given a list of indices, returns a list of indices, all shifted by ``shift``.

        Parameters
        ----------
        indices : list
        shift : int

        Returns
        -------
        list

        """
        return [index + shift for index in indices]

    def group_indices(self, indices, group_size):
        """Returns a list of the elements of ``indices`` grouped into tuples of size ``group_size``.

        Parameters
        ----------
        indices : list
        group_size : int

        Returns
        -------
        list

        """
        it = [iter(indices)] * group_size
        return list(zip(*it))

    @classmethod
    def validate_faces(cls, faces):
        """Raises an exception if not all faces in ``faces`` are defining either all triangles, lines
        or points.

        Parameters
        ----------
        faces : list

        Returns
        -------

        """
        if not faces:
            return
        if len(faces[0]) > 3:
            raise Exception('Invalid mesh. Expected mesh composed of points, lines xor triangles.')
        for face in faces:
            if len(face) != len(faces[0]):
                # This restriction could be removed by splitting into multiple primitives.
                raise NotImplementedError('Invalid mesh. Expected mesh composed of points, lines xor triangles.')

    @classmethod
    def validate_vertices(cls, vertices):
        """Raise an exception if there are either too many vertices, or the vertices do not
        represent points in 3-space.

        Parameters
        ----------
        vertices : list

        Returns
        -------

        """
        if len(vertices) > 4294967295:
            # This restriction could be removed by splitting into multiple primitives.
            raise Exception('Invalid mesh.  Too many vertices.')
        positions = list(vertices.values()) if isinstance(vertices, dict) else vertices
        for position in positions:
            if len(position) != 3:
                raise Exception('Invalid mesh.  Vertices are expected to be points in 3-space.')

    @classmethod
    def from_vertices_and_faces(cls, context, vertices, faces, mesh_name=None, extras=None):
        """Construct a :class:`compas.files.GLTFMesh` object from lists of vertices and faces.
        Vertices can be given as either a list of xyz-tuples or -lists, in which case
        the faces reference vertices by index, or vertices can be given as a dictionary of
        key-value pairs where the values are xyz-tuples or -lists and the faces reference the keys.

        Parameters
        ----------
        context : :class:`compas.files.GLTFContent`
        vertices : Union[list, dict]
        faces : list
        mesh_name : str
        extras : object

        Returns
        -------
        :class:`compas.files.GLTFMesh`

        """
        cls.validate_faces(faces)
        cls.validate_vertices(vertices)
        mode = get_mode(faces)
        if isinstance(vertices, dict):
            index_by_key = {}
            positions = []
            for key, position in vertices.items():
                positions.append(position)
                index_by_key[key] = len(positions) - 1
            face_list = [index_by_key[key] for key in itertools.chain(*faces)]
        else:
            positions = vertices
            face_list = list(itertools.chain(*faces))

        primitive = PrimitiveData({'POSITION': positions}, face_list, None, mode, None, None)

        return cls([primitive], context, mesh_name=mesh_name, extras=extras)

    @classmethod
    def from_mesh(cls, context, mesh):
        """Construct a :class:`compas.files.GLTFMesh` object from a compas mesh.

        Parameters
        ----------
        context : :class:`compas.files.GLTFContent`
        mesh : :class:`compas.datastructures.Mesh`

        Returns
        -------
        :class:`compas.files.GLTFMesh`
        """
        vertices, faces = mesh.to_vertices_and_faces()
        return cls.from_vertices_and_faces(context, vertices, faces)

    def to_data(self, primitives):
        """Returns a JSONable dictionary object in accordance with glTF specifications.

        Parameters
        ----------
        primitives : list

        Returns
        -------
        dict
        """
        mesh_dict = {'primitives': primitives}
        if self.mesh_name is not None:
            mesh_dict['name'] = self.mesh_name
        if self.weights is not None:
            mesh_dict['weights'] = self.weights
        if self.extras is not None:
            mesh_dict['extras'] = self.extras
        if self.extensions is not None:
            mesh_dict['extensions'] = self.extensions
        return mesh_dict

    @classmethod
    def from_data(cls, mesh, context, primitive_data_list):
        """Creates a :class:`compas.files.GLTFMesh` from a glTF node dictionary
        and inserts it in the provided context.

        Parameters
        ----------
        mesh : dict
        context : :class:`compas.files.GLTFContent`
        primitive_data_list : list

        Returns
        -------
        :class:`compas.files.GLTFMesh`
        """
        if mesh is None:
            return None
        return cls(
            primitive_data_list=primitive_data_list,
            context=context,
            mesh_name=mesh.get('name'),
            weights=mesh.get('weights'),
            extras=mesh.get('extras'),
            extensions=mesh.get('extensions'),
        )



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import array
import base64
import json
import os
import struct

from compas.files.gltf.constants import COMPONENT_TYPE_ENUM
from compas.files.gltf.constants import COMPONENT_TYPE_FLOAT
from compas.files.gltf.constants import COMPONENT_TYPE_UNSIGNED_INT
from compas.files.gltf.constants import NUM_COMPONENTS_BY_TYPE_ENUM
from compas.files.gltf.constants import TYPE_MAT4
from compas.files.gltf.constants import TYPE_SCALAR
from compas.files.gltf.constants import TYPE_VEC2
from compas.files.gltf.constants import TYPE_VEC3
from compas.files.gltf.constants import TYPE_VEC4


# This fails on IronPython 2.7.8 (eg. Rhino 6 on Windows)
# but works on IronPython 2.7.9 (Rhino 6 on Mac)
try:
    struct.pack_into('<I', bytearray(4), 0, 0)
    USE_BYTEARRAY_BUFFERS = True
except TypeError:
    USE_BYTEARRAY_BUFFERS = False


class GLTFExporter(object):
    """Export a glTF or glb file based on the supplied scene and ancillary data.

    Parameters
    ----------
    filepath : str
        Location where the glTF or glb is to be written. The extension of the filepath
        determines which format will be used. If there will be an accompanying binary file,
        it will be written in the same directory.
    content : :class:`compas.files.GLTFContent`
    embed_data : bool
        When ``True``, all mesh and other data will be embedded as data uri's in the glTF json,
        with the exception of external image data.
        When ``False``, the data will be written to an external binary file or chunk.

    """

    def __init__(self, filepath, content, embed_data=False):
        self.gltf_filepath = filepath
        self._dirname = None
        self._filename = None
        self._ext = None

        self._embed_data = embed_data
        self._content = content

        self._gltf_dict = {}
        self._mesh_index_by_key = {}
        self._node_index_by_key = {}
        self._scene_index_by_key = {}
        self._camera_index_by_key = {}
        self._skin_index_by_key = {}
        self._material_index_by_key = {}
        self._texture_index_by_key = {}
        self._sampler_index_by_key = {}
        self._image_index_by_key = {}
        self._buffer = b''

        self.load()

    @property
    def embed_data(self):
        return self._embed_data

    @embed_data.setter
    def embed_data(self, value):
        if value != self._embed_data:
            self._embed_data = value
            self.load()

    def load(self):
        """Creates the json object and the binary data (if any) to be written.

        Returns
        -------
        None

        """
        self._content.remove_orphans()
        self._content.check_if_forest()

        self._set_initial_gltf_dict()
        self._mesh_index_by_key = self._get_index_by_key(self._content.meshes)
        self._node_index_by_key = self._get_index_by_key(self._content.nodes)
        self._scene_index_by_key = self._get_index_by_key(self._content.scenes)
        self._camera_index_by_key = self._get_index_by_key(self._content.cameras)
        self._skin_index_by_key = self._get_index_by_key(self._content.skins)
        self._material_index_by_key = self._get_index_by_key(self._content.materials)
        self._texture_index_by_key = self._get_index_by_key(self._content.textures)
        self._sampler_index_by_key = self._get_index_by_key(self._content.samplers)
        self._image_index_by_key = self._get_index_by_key(self._content.images)
        self._buffer = b''

        self._set_path_attributes()
        self._add_meshes()
        self._add_nodes()
        self._add_scenes()
        self._add_cameras()
        self._add_skins()
        self._add_materials()
        self._add_textures()
        self._add_samplers()
        self._add_images()
        self._add_animations()
        self._add_buffer()

    def _get_index_by_key(self, d):
        return {key: index for index, key in enumerate(d)}

    def export(self):
        """Writes the json to *.gltf* or *.glb*, and binary data to *.bin* as required.

        Returns
        -------
        None

        """
        gltf_json = json.dumps(self._gltf_dict, indent=4)

        if self._ext == '.gltf':
            with open(self.gltf_filepath, 'w') as f:
                f.write(gltf_json)
            if not self._embed_data and len(self._buffer) > 0:
                with open(self.get_bin_path(), 'wb') as f:
                    f.write(self._buffer)

        if self._ext == '.glb':
            with open(self.gltf_filepath, 'wb') as f:
                gltf_data = gltf_json.encode()

                length_gltf = len(gltf_data)
                spaces_gltf = (4 - (length_gltf & 3)) & 3
                length_gltf += spaces_gltf

                length_bin = len(self._buffer)
                zeros_bin = (4 - (length_bin & 3)) & 3
                length_bin += zeros_bin

                length = 12 + 8 + length_gltf
                if length_bin > 0:
                    length += 8 + length_bin

                f.write('glTF'.encode('ascii'))
                f.write(struct.pack('<I', 2))
                f.write(struct.pack('<I', length))

                f.write(struct.pack('<I', length_gltf))
                f.write('JSON'.encode('ascii'))
                f.write(gltf_data)
                for i in range(0, spaces_gltf):
                    f.write(' '.encode())

                if length_bin > 0:
                    f.write(struct.pack('<I', length_bin))
                    f.write('BIN\0'.encode())
                    f.write(self._buffer)
                    for i in range(0, zeros_bin):
                        f.write('\0'.encode())

    def _add_images(self):
        if not self._content.images:
            return
        images_list = [None] * len(self._content.images)
        for key, image_data in self._content.images.items():
            uri = self._construct_image_data_uri(image_data) if self.embed_data else None
            buffer_view = self._construct_buffer_view(image_data.data) if not self.embed_data else None
            images_list[self._image_index_by_key[key]] = image_data.to_data(uri, buffer_view)
        self._gltf_dict['images'] = images_list

    def _construct_image_data_uri(self, image_data):
        if image_data.data is None:
            return None
        return (
            'data:'
            + (image_data.mime_type if image_data.mime_type else '')
            + ';base64,' + base64.b64encode(image_data.data).decode('ascii')
        )

    def _add_samplers(self):
        if not self._content.samplers:
            return
        samplers_list = [None] * len(self._content.samplers)
        for key, sampler_data in self._content.samplers.items():
            samplers_list[self._sampler_index_by_key[key]] = sampler_data.to_data()
        self._gltf_dict['samplers'] = samplers_list

    def _add_textures(self):
        if not self._content.textures:
            return
        textures_list = [None] * len(self._content.textures)
        for key, texture_data in self._content.textures.items():
            textures_list[self._texture_index_by_key[key]] = texture_data.to_data(self._sampler_index_by_key, self._image_index_by_key)
        self._gltf_dict['textures'] = textures_list

    def _add_materials(self):
        if not self._content.materials:
            return
        materials_list = [None] * len(self._content.materials)
        for key, material_data in self._content.materials.items():
            materials_list[self._material_index_by_key[key]] = material_data.to_data(self._texture_index_by_key)
        self._gltf_dict['materials'] = materials_list

    def _add_skins(self):
        if not self._content.skins:
            return
        skins_list = [None] * len(self._content.skins)
        for key, skin_data in self._content.skins.items():
            accessor_index = self._construct_accessor(skin_data.inverse_bind_matrices, COMPONENT_TYPE_FLOAT, TYPE_MAT4)
            skins_list[self._skin_index_by_key[key]] = skin_data.to_data(self._node_index_by_key, accessor_index)
        self._gltf_dict['skins'] = skins_list

    def _add_cameras(self):
        if not self._content.cameras:
            return
        camera_list = [None] * len(self._content.cameras)
        for key, camera_data in self._content.cameras.items():
            camera_list[self._camera_index_by_key[key]] = camera_data.to_data()
        self._gltf_dict['cameras'] = camera_list

    def _add_meshes(self):
        if not self._content.meshes:
            return
        mesh_list = [None] * len(self._content.meshes)
        for key, mesh_data in self._content.meshes.items():
            primitives = self._construct_primitives(mesh_data)
            mesh_list[self._mesh_index_by_key[key]] = mesh_data.to_data(primitives)
        self._gltf_dict['meshes'] = mesh_list

    def _add_buffer(self):
        if not self._buffer:
            return
        buffer = {'byteLength': len(self._buffer)}
        if self._embed_data:
            buffer['uri'] = 'data:application/octet-stream;base64,' + base64.b64encode(self._buffer).decode('ascii')
        elif self._ext == '.gltf':
            buffer['uri'] = self.get_bin_filename()
        self._gltf_dict['buffers'] = [buffer]

    def _add_animations(self):
        if not self._content.animations:
            return None
        animation_list = []
        for animation_data in self._content.animations.values():
            samplers_list = self._construct_animation_samplers_list(animation_data)
            animation_list.append(animation_data.to_data(samplers_list, self._node_index_by_key))
        self._gltf_dict['animations'] = animation_list

    def _construct_animation_samplers_list(self, animation_data):
        sampler_index_by_key = animation_data.get_sampler_index_by_key()
        samplers_list = [None] * len(sampler_index_by_key)
        for key, sampler_data in animation_data.samplers_dict.items():
            input_accessor = self._construct_accessor(sampler_data.input, COMPONENT_TYPE_FLOAT, TYPE_SCALAR, include_bounds=True)
            type_ = TYPE_VEC3
            if isinstance(sampler_data.output[0], int) or isinstance(sampler_data.output[0], float):
                type_ = TYPE_SCALAR
            elif len(sampler_data.output[0]) == 4:
                type_ = TYPE_VEC4
            output_accessor = self._construct_accessor(sampler_data.output, COMPONENT_TYPE_FLOAT, type_)
            samplers_list[sampler_index_by_key[key]] = sampler_data.to_data(input_accessor, output_accessor)
        return samplers_list

    def _set_initial_gltf_dict(self):
        asset_dict = {'version': '2.0'}
        gltf_dict = {'asset': asset_dict}
        if self._content.extras:
            gltf_dict['extras'] = self._content.extras
        if self._content.extensions:
            gltf_dict['extensions'] = self._content.extensions
        self._gltf_dict = gltf_dict

    def _add_scenes(self):
        if not self._content.scenes:
            return
        if self._content.default_scene_key is not None:
            self._gltf_dict['scene'] = self._scene_index_by_key[self._content.default_scene_key]
        scene_list = [None] * len(self._content.scenes.values())
        for key, scene in self._content.scenes.items():
            scene_list[self._scene_index_by_key[key]] = scene.to_data(self._node_index_by_key)
        self._gltf_dict['scenes'] = scene_list

    def _add_nodes(self):
        if not self._content.nodes:
            return
        node_list = [None] * len(self._content.nodes)
        for key, node in self._content.nodes.items():
            node_list[self._node_index_by_key[key]] = node.to_data(
                self._node_index_by_key,
                self._mesh_index_by_key,
                self._camera_index_by_key,
                self._skin_index_by_key,
            )
        self._gltf_dict['nodes'] = node_list

    def _construct_primitives(self, mesh_data):
        primitives = []
        for primitive_data in mesh_data.primitive_data_list:
            indices_accessor = self._construct_accessor(primitive_data.indices, COMPONENT_TYPE_UNSIGNED_INT, TYPE_SCALAR)

            attributes = {}
            for attr in primitive_data.attributes:
                component_type = COMPONENT_TYPE_UNSIGNED_INT if attr.startswith('JOINT') else COMPONENT_TYPE_FLOAT
                type_ = TYPE_VEC3
                if len(primitive_data.attributes[attr][0]) == 4:
                    type_ = TYPE_VEC4
                if len(primitive_data.attributes[attr][0]) == 2:
                    type_ = TYPE_VEC2
                attributes[attr] = self._construct_accessor(primitive_data.attributes[attr], component_type, type_, True)

            targets = []
            for target in primitive_data.targets or []:
                target_dict = {}
                for attr in target:
                    component_type = COMPONENT_TYPE_FLOAT
                    type_ = TYPE_VEC3
                    target_dict[attr] = self._construct_accessor(target[attr], component_type, type_, True)
                targets.append(target_dict)

            primitive_dict = primitive_data.to_data(indices_accessor, attributes, targets, self._material_index_by_key)

            primitives.append(primitive_dict)
        return primitives

    def _construct_accessor(self, data, component_type, type_, include_bounds=False):
        if data is None:
            return None
        count = len(data)

        fmt_char = COMPONENT_TYPE_ENUM[component_type]
        fmt = '<' + fmt_char * NUM_COMPONENTS_BY_TYPE_ENUM[type_]

        component_size = struct.calcsize('<' + fmt_char)
        if component_type == 'MAT2' and component_size == 1:
            fmt = '<FFxxFFxx'.replace('F', fmt_char)
        elif component_type == 'MAT3' and component_size == 1:
            fmt = '<FFFxFFFxFFFx'.replace('F', fmt_char)
        elif component_type == 'MAT3' and component_size == 2:
            fmt = '<FFFxxFFFxxFFFxx'.replace('F', fmt_char)

        component_len = struct.calcsize(fmt)

        # ensure bytes_ length is divisible by 4
        size = count * component_len
        size += (4 - size % 4) % 4

        if USE_BYTEARRAY_BUFFERS:
            bytes_ = bytearray(size)
        else:
            bytes_ = array.array('B', [0] * size)

        for i, datum in enumerate(data):
            if isinstance(datum, int) or isinstance(datum, float):
                struct.pack_into(fmt, bytes_, (i * component_len), datum)
            else:
                struct.pack_into(fmt, bytes_, (i * component_len), *datum)

        buffer_view_index = self._construct_buffer_view(bytes_)
        accessor_dict = {
            'bufferView': buffer_view_index,
            'count': count,
            'componentType': component_type,
            'type': type_,
        }
        if include_bounds:
            try:
                # Here we check if ``data`` contains tuples,
                # and compute min/max per coordinate.
                _ = [e for e in data[0]]
                minimum = tuple(map(min, zip(*data)))
                maximum = tuple(map(max, zip(*data)))
            except TypeError:
                # Here, ``data`` must contain primitives and not tuples,
                # so min and max are more simply computed.
                minimum = (min(data),)
                maximum = (max(data),)
            accessor_dict['min'] = minimum
            accessor_dict['max'] = maximum

        self._gltf_dict.setdefault('accessors', []).append(accessor_dict)

        return len(self._gltf_dict['accessors']) - 1

    def _construct_buffer_view(self, bytes_):
        if not bytes_:
            return None
        byte_offset = self._update_buffer(bytes_)
        buffer_view_dict = {
            'buffer': 0,
            'byteLength': len(bytes_),
            'byteOffset': byte_offset,
        }

        self._gltf_dict.setdefault('bufferViews', []).append(buffer_view_dict)

        return len(self._gltf_dict['bufferViews']) - 1

    def _update_buffer(self, bytes_):
        byte_offset = len(self._buffer)
        # If bytes_ was not created as bytearray, cast now
        if not USE_BYTEARRAY_BUFFERS:
            bytes_ = bytearray(bytes_)
        self._buffer += bytes_
        return byte_offset

    def _set_path_attributes(self):
        dirname, basename = os.path.split(self.gltf_filepath)
        root, ext = os.path.splitext(basename)
        self._dirname = dirname
        self._filename = root
        self._ext = ext.lower()

    def get_bin_path(self):
        return os.path.join(self._dirname, self._filename + '.bin')

    def get_bin_filename(self):
        return self._filename + '.bin'



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from subprocess import Popen
import sys

from functools import wraps

from numpy import array
from numpy import asarray
from numpy import atleast_2d
from numpy import nan_to_num
from numpy import nonzero
from numpy import seterr
from numpy import sum
from numpy import absolute
from numpy import cross
from numpy.linalg import cond

from scipy.linalg import cho_factor
from scipy.linalg import cho_solve
from scipy.linalg import lstsq
from scipy.linalg import qr
from scipy.linalg import svd
from scipy.io import loadmat
from scipy.io import savemat
from scipy.sparse.linalg import factorized
from scipy.sparse.linalg import spsolve

import compas

old_settings = seterr(all='ignore')


__all__ = [
    'nullspace',
    'rank',
    'dof',
    'pivots',
    'nonpivots',
    'rref',
    'rref_sympy',
    'rref_matlab',
    'uvw_lengths',
    'normrow',
    'normalizerow',
    'rot90',
    'solve_with_known',
    'spsolve_with_known',
    'chofactor',
    'lufactorized'
]


# ==============================================================================
# Fundamentals
# ==============================================================================

def nullspace(A, tol=0.001):
    r"""Calculates the nullspace of the input matrix A.

    Parameters
    ----------
    A : array-like
        Matrix A represented as an array or list.
    tol : float
        Tolerance.

    Returns
    -------
    array
        Null(A).

    Notes
    -----
    The nullspace is the set of vector solutions to the equation

    .. math::

        \mathbf{A} \mathbf{x} = 0

    where 0 is a vector of zeros.

    When determining the nullspace using SVD decomposition (A = U S Vh),
    the right-singular vectors (rows of Vh or columns of V) corresponding to
    vanishing singular values of A, span the nullspace of A.

    Examples
    --------
    >>> nullspace(array([[2, 3, 5], [-4, 2, 3]]))
    array([[-0.03273853],
           [-0.85120179],
           [ 0.52381648]])

    """
    A = atleast_2d(asarray(A, dtype=float))
    u, s, vh = svd(A, compute_uv=True)
    tol = s[0] * tol
    r = (s >= tol).sum()
    # nullspace
    # ---------
    # if A is m x n
    # the last (n - r) columns of v (or the last n - r rows of vh)
    null = vh[r:].conj().T
    return null


def rank(A, tol=0.001):
    r"""Calculates the rank of the input matrix A.

    Parameters
    ----------
    A : array-like
        Matrix A represented as an array or list.
    tol : float
        Tolerance.

    Returns
    -------
    int
        rank(A)

    Notes
    -----
    The rank of a matrix is the maximum number of linearly independent rows in
    a matrix. Note that the row rank is equal to the column rank of the matrix.

    Examples
    --------
    >>> rank([[1, 2, 1], [-2, -3, 1], [3, 5, 0]])
    2

    """
    A = atleast_2d(asarray(A, dtype=float))
    s = svd(A, compute_uv=False)
    tol = s[0] * tol
    r = (s >= tol).sum()
    return r


def dof(A, tol=0.001, condition=False):
    r"""Returns the degrees-of-freedom of the input matrix A.

    Parameters
    ----------
    A : array-like
        Matrix A represented as an array or list.
    tol : float (0.001)
        Tolerance.
    condition : bool (False)
        Return the condition number of the matrix.

    Returns
    -------
    int
        Column degrees-of-freedom.
    int
        Row degrees-of-freedom.
    float
        Condition number, if ``condition`` is ``True``.

    Notes
    -----
    The degrees-of-freedom are the number of columns and rows minus the rank.

    Examples
    --------
    >>> dof([[2, -1, 3], [1, 0, 1], [0, 2, -1], [1, 1, 4]], condition=True)
    (0, 1, 5.073596551276727)

    """
    A = atleast_2d(asarray(A, dtype=float))
    r = rank(A, tol=tol)
    k = A.shape[1] - r
    m = A.shape[0] - r
    if condition:
        c = cond(A)
        return k, m, c
    return k, m


def pivots(U, tol=None):
    r"""Identify the pivots of input matrix U.

    Parameters
    ----------
    U : array-like
        Matrix U represented as an array or list.

    Returns
    -------
    list
        Pivot column indices.

    Notes
    -----
    If the matrix U is in Reduced Row Echelon Form,
    the pivots are the columns with leading non-zero coefficients per row.

    Examples
    --------
    >>> A = [[1, 0, 1, 3], [2, 3, 4, 7], [-1, -3, -3, -4]]
    >>> n = rref_sympy(A)
    >>> pivots(n)
    [0, 1]

    """
    if tol is None:
        tol = sys.float_info.epsilon
    U = atleast_2d(array(U, dtype=float))
    U[absolute(U) < tol] = 0.0
    pivots = []
    for row in U:
        cols = nonzero(row)[0]
        if len(cols):
            pivots.append(cols[0])
    return pivots


def nonpivots(U, tol=None):
    r"""Identify the non-pivots of input matrix U.

    Parameters
    ----------
    U : array-like
        Matrix U represented as an array or list.

    Returns
    -------
    list
        Non-pivot column indices.

    Notes
    -----
    If the matrix U is in Reduced Row Echelon Form,
    the nonpivots are the columns with non-zero coefficients that are not leading their row.

    Examples
    --------
    >>> A = [[1, 0, 1, 3], [2, 3, 4, 7], [-1, -3, -3, -4]]
    >>> n = rref_sympy(A)
    >>> nonpivots(n)
    [2, 3]

    """
    U = atleast_2d(asarray(U, dtype=float))
    cols = pivots(U, tol=tol)
    return list(set(range(U.shape[1])) - set(cols))


def rref(A, tol=None):
    r"""Reduced row-echelon form of matrix A.

    Parameters
    ----------
    A : array-like
        Matrix A represented as an array or list.
    tol : float
        Tolerance.

    Returns
    -------
    array
        RREF of A.

    Notes
    -----
    A matrix is in reduced row-echelon form after Gauss-Jordan elimination.

    Examples
    --------
    >>>

    """
    A = atleast_2d(asarray(A, dtype=float))

    # do qr with column pivoting
    # to have non-decreasing absolute values on the diagonal of R
    # column pivoting ensures that the largest absolute value is used
    # as leading element
    _, U = qr(A)
    lead_pos = 0
    num_rows, num_cols = U.shape
    for r in range(num_rows):
        if lead_pos >= num_cols:
            return
        i = r
        # find a nonzero lead in column lead_pos
        while U[i][lead_pos] == 0:
            i += 1
            if i == num_rows:
                i = r
                lead_pos += 1
                if lead_pos == num_cols:
                    return
        # swap the row with the nonzero lead with the current row
        U[[i, r]] = U[[r, i]]
        # "normalize" the values of the row
        lead_val = U[r][lead_pos]
        U[r] = U[r] / lead_val
        # make sure all other column values are zero
        for i in range(num_rows):
            if i != r:
                lead_val = U[i][lead_pos]
                U[i] = U[i] - lead_val * U[r]
        # go to the next column
        lead_pos += 1
    return U


def rref_sympy(A, tol=None):
    r"""Reduced row-echelon form of matrix A.

    Parameters
    ----------
    A : array-like
        Matrix A represented as an array or list.
    tol : float
        Tolerance.

    Returns
    -------
    array
        RREF of A.

    Notes
    -----
    A matrix is in reduced row-echelon form after Gauss-Jordan elimination, the
    result is independent of the method/algorithm used.

    Examples
    --------
    >>> A = [[1, 0, 1, 3], [2, 3, 4, 7], [-1, -3, -3, -4]]
    >>> n = rref_sympy(A)
    >>> array(n)
    array([[1, 0, 1.00000000000000, 3.00000000000000],
           [0, 1, 0.666666666666667, 0.333333333333333],
           [0, 0, 0, 0]], dtype=object)

    """
    import sympy

    A = atleast_2d(asarray(A, dtype=float))
    return sympy.Matrix(A).rref()[0].tolist()


def rref_matlab(A, ifile, ofile, tol=None):
    r"""Reduced row-echelon form of matrix A.

    Parameters
    ----------
    A : array-like
        Matrix A represented as an array or list.
    tol : float
        Tolerance.

    Returns
    -------
    array
        RREF of A.

    Notes
    -----
    A matrix is in reduced row-echelon form after Gauss-Jordan elimination, the
    result is independent of the method/algorithm used.

    Examples
    --------
    >>>

    """
    A = atleast_2d(asarray(A, dtype=float))

    idict = {'A': A}
    savemat(ifile, idict)

    matlab = ['matlab']
    if compas.is_windows():
        options = ['-nosplash', '-wait', '-r']
    else:
        options = ['-nosplash', '-r']
    command = ["load('{0}');[R, jb]=rref(A);save('{1}');exit;".format(ifile, ofile)]

    p = Popen(matlab + options + command)
    stdout, stderr = p.communicate()

    odict = loadmat(ofile)
    return odict['R']


# ==============================================================================
# Factorisation
# ==============================================================================

class Memoized:
    """"""

    def __init__(self, f):
        self.f = f
        self.memo = {}

    def __call__(self, *args):
        key = args[-1]
        if key in self.memo:
            return self.memo[key]
        self.memo[key] = res = self.f(args[0])
        return res


def memoize(f):
    memo = {}

    @wraps(f)
    def wrapper(*args):
        key = args[-1]
        if key in memo:
            return memo[key]
        memo[key] = res = f(args[0])
        return res

    return wrapper


def _chofactor(A):
    r"""Returns the Cholesky factorisation/decomposition matrix.

    Parameters
    ----------
    A : array
        Matrix A represented as an (m x m) array.

    Returns
    -------
    array
        Matrix (m x m) with upper/lower triangle containing Cholesky factor of A.

    Notes
    -----
    The Cholesky factorisation decomposes a Hermitian positive-definite matrix
    into the product of a lower/upper triangular matrix and its transpose.

    .. math::

        \mathbf{A} = \mathbf{L} \mathbf{L}^{\mathrm{T}}

    Examples
    --------
    >>> _chofactor(array([[25, 15, -5], [15, 18, 0], [-5, 0, 11]]))
    (array([[ 5.,  3., -1.],
           [15.,  3.,  1.],
           [-5.,  0.,  3.]]), False)

    """
    return cho_factor(A)


def _lufactorized(A):
    r"""Return a function for solving a sparse linear system (LU decomposition).

    Parameters
    ----------
    A : array
        Matrix A represented as an (m x n) array.

    Returns
    -------
    callable
        Function to solve linear system with input matrix (n x 1).

    Notes
    -----
    LU decomposition factors a matrix as the product of a lower triangular and
    an upper triangular matrix L and U.

    .. math::

        \mathbf{A} = \mathbf{L} \mathbf{U}

    Examples
    --------
    >>> fn = _lufactorized(array([[3, 2, -1], [2, -2, 4], [-1, 0.5, -1]]))
    >>> fn(array([1, -2, 0]))
    array([ 1., -2., -2.])

    """
    return factorized(A)


chofactor = memoize(_chofactor)
lufactorized = memoize(_lufactorized)


# ------------------------------------------------------------------------------
# Geometry
# ------------------------------------------------------------------------------

def uvw_lengths(C, X):
    r"""Calculates the lengths and co-ordinate differences.

    Parameters
    ----------
    C : sparse
        Connectivity matrix (m x n).
    X : array
        Co-ordinates of vertices/points (n x 3).

    Returns
    -------
    array
        Vectors of co-ordinate differences in x, y and z (m x 3).
    array
        Lengths of members (m x 1).

    Examples
    --------
    >>> from compas.numerical import connectivity_matrix
    >>> C = connectivity_matrix([[0, 1], [1, 2]], 'csr')
    >>> X = array([[0, 0, 0], [1, 1, 0], [0, 0, 1]])
    >>> uvw_lengths(C, X)
    (array([[ 1.,  1.,  0.],
           [-1., -1.,  1.]]), array([[1.41421356],
           [1.73205081]]))

    """
    uvw = C.dot(X)
    return uvw, normrow(uvw)


def normrow(A):
    """Calculates the 2-norm of each row of matrix A.

    Parameters
    ----------
    A : array
        Matrix A represented as an (m x n) array.

    Returns
    -------
    array
        Column vector (m x 1) of values.

    Notes
    -----
    The calculation is the Euclidean 2-norm, i.e. the square root of the sum
    of the squares of the elements in each row, this equates to the "length" of
    the m row vectors.

    Examples
    --------
    >>> normrow(array([[2, -1, 3,], [1, 0, 1], [0, 2, -1]]))
    array([[3.74165739],
           [1.41421356],
           [2.23606798]])

    """
    A = atleast_2d(asarray(A, dtype=float))
    return (sum(A ** 2, axis=1) ** 0.5).reshape((-1, 1))


def normalizerow(A, do_nan_to_num=True):
    """Normalise the rows of matrix A.

    Parameters
    ----------
    A : array
        Matrix A represented as an (m x n) array.
    do_nan_to_num : bool
        Convert NaNs and INF to numbers, default=True.

    Returns
    -------
    array
        Matrix of normalized row vectors (m x n).

    Notes
    -----
    Normalises the row vectors of A by the normrows, i.e. creates an array of
    vectors where the row vectors have length of unity.

    Tiling is not necessary, because of NumPy's broadcasting behaviour.

    Examples
    --------
    >>> normalizerow(array([[2, -1, 3,], [1, 0, 1], [0, 2, -1]]))
    array([[ 0.53452248, -0.26726124,  0.80178373],
           [ 0.70710678,  0.        ,  0.70710678],
           [ 0.        ,  0.89442719, -0.4472136 ]])

    """
    if do_nan_to_num:
        return nan_to_num(A / normrow(A))
    else:
        return A / normrow(A)


def rot90(vectors, axes):
    """Rotate an array of vectors through 90 degrees around an array of axes.

    Parameters
    ----------
    vectors : array
        An array of row vectors (m x 3).
    axes : array
        An array of axes (m x 3).

    Returns
    -------
    array
        Matrix of row vectors (m x 3).

    Notes
    -----
    Computes the cross product of each row vector with its corresponding axis,
    and then rescales the resulting normal vectors to match the length of the
    original row vectors.

    Examples
    --------
    >>> vectors = array([[2, 1, 3], [2, 6, 8]])
    >>> axes = array([[7, 0, 1], [4, 4, 2]])
    >>> rot90(vectors, axes)
    array([[-0.18456235, -3.50668461,  1.29193644],
           [ 5.3748385 , -7.5247739 ,  4.2998708 ]])

    """
    return normalizerow(cross(axes, vectors)) * normrow(vectors)


# ==============================================================================
# Solving
# ==============================================================================

def solve_with_known(A, b, x, known):
    r"""Solve a system of linear equations with part of solution known.

    Parameters
    ----------
    A : array
        Coefficient matrix represented as an (m x n) array.
    b : array
        Right-hand-side represented as an (m x 1) array.
    x : array
        Unknowns/knowns represented as an (n x 1) array.
    known : list
        The indices of the known elements of ``x``.

    Returns
    -------
    array: (n x 1) vector solution.

    Notes
    -----
    Computes the solution of the system of linear equations.

    .. math::

        \mathbf{A} \mathbf{x} = \mathbf{b}

    """
    eps = 1 / sys.float_info.epsilon
    unknown = list(set(range(x.shape[0])) - set(known))
    A11 = A[unknown, :][:, unknown]
    A12 = A[unknown, :][:, known]
    b = b[unknown] - A12.dot(x[known])
    if cond(A11) < eps:
        Y = cho_solve(cho_factor(A11), b)
        x[unknown] = Y
        return x
    Y = lstsq(A11, b)
    x[unknown] = Y[0]
    return x


def spsolve_with_known(A, b, x, known):
    r"""Solve (sparse) a system of linear equations with part of solution known.

    Parameters
    ----------
    A : array
        Coefficient matrix (sparse) represented as an (m x n) array.
    b : array
        Right-hand-side represented as an (m x 1) array.
    x : array
        Unknowns/knowns represented as an (n x 1) array.
    known : list
        The indices of the known elements of ``x``.

    Returns
    -------
    array
        (n x 1) vector solution.

    Notes
    -----
    Computes the solution (using spsolve) of the system of linear equations.

    .. math::

        \mathbf{A} \mathbf{x} = \mathbf{b}

    Same function as solve_with_known, but for sparse matrix A.

    Examples
    --------
    >>> A = array([[2, 1, 3], [2, 6, 8], [6, 8, 18]])
    >>> b = array([[1], [3], [5]])
    >>> x = array([[0.3], [0], [0]])
    >>> solve_with_known(A, b, x, [0])
    array([[ 3.00000000e-01],
           [ 4.00000000e-01],
           [-6.66247443e-17]])

    """
    unknown = list(set(range(x.shape[0])) - set(known))
    A11 = A[unknown, :][:, unknown]
    A12 = A[unknown, :][:, known]
    b = b[unknown] - A12.dot(x[known])
    x[unknown] = spsolve(A11, b)
    return x



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np
from numpy import asarray
from numpy import argmin
from numpy.linalg import det
from scipy.spatial.distance import cdist
from scipy.linalg import svd
from scipy.linalg import norm

from compas.numerical import pca_numpy
from compas.numerical import normrow
from compas.geometry import Transformation
from compas.geometry import Frame
from compas.geometry import transform_points_numpy


__all__ = ['icp_numpy']


def bestfit_transform(A, B):
    n, m = A.shape
    Am = np.mean(A, axis=0)
    Bm = np.mean(B, axis=0)
    AA = A - Am
    BB = B - Bm
    # cross-covariance matrix
    C = np.dot(AA.T, BB)
    U, S, Vt = svd(C)
    # rigid rotation of the data frames
    R = np.dot(Vt.T, U.T)
    # check for RotoReflection
    if det(R) < 0:
        Vt[m-1, :] *= -1
        R = np.dot(Vt.T, U.T)
    # translation that moves data set means to same location
    # this can be done differently (by applying three transformations (T1, R, T2))
    T = Bm.T - np.dot(R, Am.T)
    X = np.identity(m+1)
    X[:m, :m] = R
    X[:m, m] = T
    return X


def icp_numpy(source, target, tol=1e-3):
    """Align two point clouds using the Iterative Closest Point (ICP) method.

    Parameters
    ----------
    source : array_like[point]
        The source data.
    target : array_like[point]
        The target data.
    tol : float, optional
        Tolerance for finding matches.

    Returns
    -------
    ndarray[float](N, 3)
        The transformed points.
    ndarray[float](4, 4)
        The bestfit transformation matrix.

    Notes
    -----
    First we align the source with the target cloud using the frames resulting
    from a PCA of each of the clouds, simply by calculating a frame-to-frame transformation.

    This initial alignment is used to establish an initial correspondence between
    the points of the two clouds.

    Then we iteratively improve the alignment by computing successive "best-fit"
    transformations using SVD of the cross-covariance matrix of the two data sets.
    During this iterative process, we continuously update the correspondence
    between the point clouds by finding the closest point in the target to each
    of the source points.

    The algorithm terminates when the alignment error is below a specified tolerance.

    Examples
    --------
    >>>

    """
    A = asarray(source)
    B = asarray(target)

    origin, axes, _ = pca_numpy(A)
    A_frame = Frame(origin, axes[0], axes[1])

    origin, axes, _ = pca_numpy(B)
    B_frame = Frame(origin, axes[0], axes[1])

    X = Transformation.from_frame_to_frame(A_frame, B_frame)
    A = transform_points_numpy(A, X)

    for i in range(20):
        D = cdist(A, B, 'euclidean')
        closest = argmin(D, axis=1)
        if norm(normrow(A - B[closest])) < tol:
            break
        X = bestfit_transform(A, B[closest])
        A = transform_points_numpy(A, X)

    return A, X



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from itertools import groupby
from compas.geometry import Point

from compas.geometry import NurbsCurve

from compas_rhino.conversions import point_to_rhino
from compas_rhino.conversions import point_to_compas
# from compas_rhino.conversions import circle_to_rhino
# from compas_rhino.conversions import ellipse_to_rhino
from compas_rhino.conversions import line_to_rhino

from .curve import RhinoCurve

import Rhino.Geometry


def rhino_curve_from_parameters(points, weights, knots, multiplicities, degree):
    rhino_curve = Rhino.Geometry.NurbsCurve(3, True, degree + 1, len(points))
    for index, (point, weight) in enumerate(zip(points, weights)):
        rhino_curve.Points.SetPoint(index, point_to_rhino(point), weight)
    knotvector = [knot for knot, mult in zip(knots, multiplicities) for _ in range(mult)]
    # account for superfluous knots
    # https://developer.rhino3d.com/guides/opennurbs/superfluous-knots/
    p = len(points)
    o = degree + 1
    k = p + o
    if len(knotvector) == k:
        knotvector[:] = knotvector[1:-1]
    for index, knot in enumerate(knotvector):
        rhino_curve.Knots[index] = knot
    return rhino_curve


class RhinoNurbsCurve(NurbsCurve, RhinoCurve):
    """Class representing a NURBS curve based on the NurbsCurve of Rhino.Geometry.

    Parameters
    ----------
    name : str, optional
        Name of the curve.

    Attributes
    ----------
    points : list[:class:`compas.geometry.Point`], read-only
        The control points of the curve.
    weights : list[float], read-only
        The weights of the control points.
    knots : list[float], read-only
        The knot vector, without duplicates.
    multiplicities : list[int], read-only
        The multiplicities of the knots in the knot vector.
    knotsequence : list[float], read-only
        The knot vector, with repeating values according to the multiplicities.
    continuity : int, read-only
        The degree of continuity of the curve.
    degree : int, read-only
        The degree of the curve.
    order : int, read-only
        The order of the curve (degree + 1).
    is_rational : bool, read-only
        True is the curve is rational.

    References
    ----------
    * https://developer.rhino3d.com/api/RhinoCommon/html/T_Rhino_Geometry_NurbsCurve.htm
    * https://en.wikipedia.org/wiki/Non-uniform_rational_B-spline
    * https://developer.rhino3d.com/guides/opennurbs/nurbs-geometry-overview/

    """

    def __init__(self, name=None):
        super(RhinoNurbsCurve, self).__init__(name=name)
        self.rhino_curve = None

    # ==============================================================================
    # Data
    # ==============================================================================

    @property
    def data(self):
        # add superfluous knots
        # for compatibility with all/most other NURBS implementations
        # https://developer.rhino3d.com/guides/opennurbs/superfluous-knots/
        multiplicities = self.multiplicities[:]
        multiplicities[0] += 1
        multiplicities[-1] += 1
        return {
            'points': [point.data for point in self.points],
            'weights': self.weights,
            'knots': self.knots,
            'multiplicities': multiplicities,
            'degree': self.degree,
            'is_periodic': self.is_periodic,
        }

    @data.setter
    def data(self, data):
        points = [Point.from_data(point) for point in data['points']]
        weights = data['weights']
        knots = data['knots']
        multiplicities = data['multiplicities']
        degree = data['degree']
        # is_periodic = data['is_periodic']
        # have not found a way to actually set this
        # not sure if that is actually possible...
        self.rhino_curve = rhino_curve_from_parameters(points, weights, knots, multiplicities, degree)

    # ==============================================================================
    # Rhino Properties
    # ==============================================================================

    # ==============================================================================
    # Properties
    # ==============================================================================

    @property
    def points(self):
        if self.rhino_curve:
            return [point_to_compas(point) for point in self.rhino_curve.Points]

    @property
    def weights(self):
        if self.rhino_curve:
            return [point.Weight for point in self.rhino_curve.Points]

    @property
    def knots(self):
        if self.rhino_curve:
            return [key for key, _ in groupby(self.rhino_curve.Knots)]

    @property
    def knotsequence(self):
        if self.rhino_curve:
            return list(self.rhino_curve.Knots)

    @property
    def multiplicities(self):
        if self.rhino_curve:
            return [len(list(group)) for _, group in groupby(self.rhino_curve.Knots)]

    @property
    def degree(self):
        if self.rhino_curve:
            return self.rhino_curve.Degree

    @property
    def order(self):
        if self.rhino_curve:
            return self.rhino_curve.Order

    @property
    def is_rational(self):
        if self.rhino_curve:
            return self.rhino_curve.IsRational

    # ==============================================================================
    # Constructors
    # ==============================================================================

    @classmethod
    def from_parameters(cls, points, weights, knots, multiplicities, degree, is_periodic=False):
        """Construct a NURBS curve from explicit curve parameters.

        Parameters
        ----------
        points : list[:class:`compas.geometry.Point`]
            The control points.
        weights : list[float]
            The control point weights.
        knots : list[float]
            The curve knots, without duplicates.
        multiplicities : list[int]
            The multiplicities of the knots.
        degree : int
            The degree of the curve.
        is_periodic : bool, optional
            Flag indicating whether the curve is periodic or not.
            Note that this parameters is currently not supported.

        Returns
        -------
        :class:`compas_rhino.geometry.RhinoNurbsCurve`

        """
        curve = cls()
        curve.rhino_curve = rhino_curve_from_parameters(points, weights, knots, multiplicities, degree)
        return curve

    @classmethod
    def from_points(cls, points, degree=3, is_periodic=False):
        """Construct a NURBS curve from control points.

        Parameters
        ----------
        points : list[:class:`compas.geometry.Point`]
            The control points.
        degree : int, optional
            The degree of the curve.
        is_periodic : bool, optional
            Flag indicating whether the curve is periodic or not.

        Returns
        -------
        :class:`compas_rhino.geometry.RhinoNurbsCurve`

        """
        points[:] = [point_to_rhino(point) for point in points]
        curve = cls()
        curve.rhino_curve = Rhino.Geometry.NurbsCurve.Create(is_periodic, degree, points)
        return curve

    @classmethod
    def from_interpolation(cls, points, precision=1e-3):
        """Construct a NURBS curve by interpolating a set of points.

        Parameters
        ----------
        points : list[:class:`compas.geometry.Point`]
            The control points.
        precision : float, optional
            The required precision of the interpolation.
            This parameter is currently not supported.

        Returns
        -------
        :class:`compas_rhino.geometry.RhinoNurbsCurve`

        """
        curve = cls()
        curve.rhino_curve = Rhino.Geometry.NurbsCurve.CreateHSpline([point_to_rhino(point) for point in points])
        return curve

    # @classmethod
    # def from_circle(cls, circle):
    #     """Construct a NURBS curve from a circle.

    #     Parameters
    #     ----------
    #     circle : :class:`compas.geometry.Circle`
    #         A circle geometry.

    #     Returns
    #     -------
    #     :class:`compas_rhino.geometry.RhinoNurbsCurve`

    #     """
    #     curve = cls()
    #     curve.rhino_curve = Rhino.Geometry.NurbsCurve.CreateFromCircle(circle_to_rhino(circle))
    #     return curve

    # @classmethod
    # def from_ellipse(cls, ellipse):
    #     """Construct a NURBS curve from an ellipse.

    #     Parameters
    #     ----------
    #     ellipse : :class:`compas.geometry.Ellipse`
    #         An ellipse geometry.

    #     Returns
    #     -------
    #     :class:`compas_rhino.geometry.RhinoNurbsCurve`

    #     """
    #     curve = cls()
    #     curve.rhino_curve = Rhino.Geometry.NurbsCurve.CreateFromEllipse(ellipse_to_rhino(ellipse))
    #     return curve

    @classmethod
    def from_line(cls, line):
        """Construct a NURBS curve from a line.

        Parameters
        ----------
        line : :class:`compas.geometry.Line`
            A line geometry.

        Returns
        -------
        :class:`compas_rhino.geometry.RhinoNurbsCurve`

        """
        curve = cls()
        curve.rhino_curve = Rhino.Geometry.NurbsCurve.CreateFromLine(line_to_rhino(line))
        return curve

    # ==============================================================================
    # Conversions
    # ==============================================================================

    # ==============================================================================
    # Methods
    # ==============================================================================



"""
Draws a frame or plane.
"""
from ghpythonlib.componentbase import executingcomponent as component

from compas_ghpython.artists import FrameArtist
from compas_ghpython.components import coerce_frame


class CompasFrame(component):
    def RunScript(self, frame):
        plane = None

        if frame:
            frame = coerce_frame(frame)
            artist = FrameArtist(frame)

            plane = artist.draw()

        return plane



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from .curvature import *  # noqa: F401 F403
from .geodistance import *  # noqa: F401 F403
from .isolines import *  # noqa: F401 F403
from .matrices import *  # noqa: F401 F403
from .parametrisation import *  # noqa: F401 F403
from .remesh import *  # noqa: F401 F403
from .slicing import *  # noqa: F401 F403


__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

try:
    basestring
except NameError:
    basestring = str

from compas_rhino.forms.base import BaseForm

from System.Windows.Forms import PictureBox
from System.Windows.Forms import PictureBoxSizeMode
from System.Windows.Forms import DockStyle
from System.Drawing import Image
from System.Net import WebClient
from System.IO import MemoryStream


__all__ = ['ImageForm', 'image_from_remote', 'image_from_local']


def image_from_remote(source):
    """Construct an image from a remote source.

    Parameters
    ----------
    source : str
        The url of the remote source.

    Returns
    -------
    System.Drawing.Image
        Representation of an image in memory.

    Examples
    --------
    .. code-block:: python

        image = image_from_remote('http://block.arch.ethz.ch/brg/images/cache/dsc02360_ni-2_cropped_1528706473_624x351.jpg')

    """
    w = WebClient()
    d = w.DownloadData(source)
    m = MemoryStream(d)
    return Image.FromStream(m)


def image_from_local(source):
    """Construct an image from a local source.

    Parameters
    ----------
    source : str
        The path to the local source file.

    Returns
    -------
    System.Drawing.Image
        Representation of an image in memory.

    Examples
    --------
    .. code-block:: python

        image = image_from_local('theblock.jpg')

    """
    return Image.FromFile(source)


class ImageForm(BaseForm):
    """A form for displaying images.

    Parameters
    ----------
    image : str or Image
        The image that should be displayed.
        This can be a url of a remote image file,
        or a local file path,
        or an instance of `System.Drawing.Image`.
    title : str, optional
        Title of the form.
    width : int, optional
        Width of the form.
    height : int, optional
        Height of the form.

    Attributes
    ----------
    image : System.Drawing.Image
        The image displayed by the form.

    Examples
    --------
    .. code-block:: python

        from compas_rhino.forms import ImageForm

        form = ImageForm('http://block.arch.ethz.ch/brg/images/cache/dsc02360_ni-2_cropped_1528706473_624x351.jpg')
        form.show()

    """

    def __init__(self, image, title='Image', width=None, height=None):
        self._image = None
        self.image = image
        super(ImageForm, self).__init__(title, width, height)

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, image):
        if isinstance(image, basestring):
            if image.startswith('http'):
                self._image = image_from_remote(image)
            else:
                self._image = image_from_local(image)
        elif isinstance(image, Image):
            self._image = image
        else:
            raise NotImplementedError

    def init(self):
        """Initialize the form.

        Returns
        -------
        None

        """
        box = PictureBox()
        box.Dock = DockStyle.Fill
        box.SizeMode = PictureBoxSizeMode.AutoSize
        box.Image = self.image
        self.image = box.Image
        self.Controls.Add(box)
        self.ClientSize = box.Size

    def on_form_closed(self, sender, e):
        """Callback for the closing event of the form.

        Parameters
        ----------
        sender : System.Object
            The sender object.
        eargs : System.Object.EventArgs
            The event arguments.

        Returns
        -------
        None

        """
        self.image.Dispose()



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

try:
    basestring
except NameError:
    basestring = str

import os
import sys
import ast

from compas_rhino.forms import TextForm
from compas_rhino.forms import ImageForm

import System

import rhinoscriptsyntax as rs
import Rhino
import clr

clr.AddReference('Rhino.UI')
import Rhino.UI  # noqa: E402
from Rhino.UI.Dialogs import ShowMessageBox  # noqa: E402

try:
    from compas_rhino.forms import PropertyListForm
except ImportError:
    from Rhino.UI.Dialogs import ShowPropertyListBox


__all__ = [
    'wait',
    'get_tolerance',
    'toggle_toolbargroup',
    'pick_point',
    'browse_for_folder',
    'browse_for_file',
    'print_display_on',
    'display_message',
    'display_text',
    'display_image',
    'display_html',
    'update_settings',
    'update_named_values',
    'screenshot_current_view',
    'select_folder',
    'select_file',
    'unload_modules',
]


# ==============================================================================
# Truly miscellaneous :)
# ==============================================================================


def screenshot_current_view(path,
                            width=1920,
                            height=1080,
                            scale=1,
                            draw_grid=False,
                            draw_world_axes=False,
                            draw_cplane_axes=False,
                            background=False):
    """Take a screenshot of the current view.

    Parameters
    ----------
    path : str
        The filepath for saving the screenshot.

    Other Parameters
    ----------------
    width : int, optional
    height : int, optional
    scale : float, optional
    draw_grid : bool, optional
    draw_world_axes : bool, optional
    draw_cplane_axes : bool, optional
    background : bool, optional

    Returns
    -------
    bool
        True if the command was successful.
        False otherwise.

    """
    properties = [draw_grid, draw_world_axes, draw_cplane_axes, background]
    properties = ["Yes" if item else "No" for item in properties]
    scale = max(1, scale)  # the rhino command requires a scale > 1
    rs.EnableRedraw(True)
    rs.Sleep(0)
    result = rs.Command("-_ViewCaptureToFile \"" + os.path.abspath(path) + "\""
                        " Width=" + str(width) +
                        " Height=" + str(height) +
                        " Scale=" + str(scale) +
                        " DrawGrid=" + properties[0] +
                        " DrawWorldAxes=" + properties[1] +
                        " DrawCPlaneAxes=" + properties[2] +
                        " TransparentBackground=" + properties[3] +
                        " _enter", False)
    rs.EnableRedraw(False)
    return result


def wait():
    """Make Rhino wait to prevent the spinning wheel from appearing.

    """
    return Rhino.RhinoApp.Wait()


def get_tolerance():
    """Get the absolute tolerance.

    Returns
    -------
    float
        The tolerance.

    """
    return rs.UnitAbsoluteTolerance()


def toggle_toolbargroup(rui, group):
    if not os.path.exists(rui) or not os.path.isfile(rui):
        return
    collection = rs.IsToolbarCollection(rui)
    if not collection:
        collection = rs.OpenToolbarCollection(rui)
        if rs.IsToolbar(collection, group, True):
            rs.ShowToolbar(collection, group)
    else:
        if rs.IsToolbar(collection, group, True):
            if rs.IsToolbarVisible(collection, group):
                rs.HideToolbar(collection, group)
            else:
                rs.ShowToolbar(collection, group)


def pick_point(message='Pick a point.'):
    point = rs.GetPoint(message)
    if point:
        return list(point)
    return None


# ==============================================================================
# File system
# ==============================================================================


def browse_for_folder(message=None, default=None):
    return rs.BrowseForFolder(folder=default, message=message, title='compas')


select_folder = browse_for_folder


def browse_for_file(title=None, folder=None, filter=None):
    if filter == 'json':
        filter = 'JSON files (*.json)|*.json||'
    elif filter == 'obj':
        filter = 'OBJ files (*.obj)|*.obj||'
    elif filter == 'fofin':
        filter = 'FOFIN session files (*.fofin)|*.fofin||'
    else:
        pass
    return rs.OpenFileName(title, filter=filter, folder=folder)


select_file = browse_for_file


# ==============================================================================
# Display
# ==============================================================================


def print_display_on(on=True):
    if on:
        rs.Command('_PrintDisplay State On Color Display Thickness 1 _Enter')
    else:
        rs.Command('_PrintDisplay State Off _Enter')


def display_message(message):
    return ShowMessageBox(message, 'Message')


def display_text(text, title='Text', width=800, height=600):
    if isinstance(text, (list, tuple)):
        text = '{0}'.format(System.Environment.NewLine).join(text)
    form = TextForm(text, title, width, height)
    return form.show()


def display_image(image, title='Image', width=800, height=600):
    form = ImageForm(image, title, width, height)
    return form.show()


def display_html():
    raise NotImplementedError


# ==============================================================================
# Settings and attributes
# ==============================================================================


def update_named_values(names, values, message='', title='Update named values', evaluate=False):
    try:
        dialog = PropertyListForm(names, values)
    except Exception:
        values = ShowPropertyListBox(message, title, names, values)
    else:
        if dialog.ShowModal(Rhino.UI.RhinoEtoApp.MainWindow):
            values = dialog.values
        else:
            values = None
    if evaluate:
        if values:
            values = list(values)
            for i in range(len(values)):
                value = values[i]
                try:
                    value = ast.literal_eval(value)
                except (TypeError, ValueError, SyntaxError):
                    pass
                values[i] = value
    return values


def update_settings(settings, message='', title='Update settings'):
    names = sorted(settings.keys())
    values = [str(settings[name]) for name in names]
    values = update_named_values(names, values, message=message, title=title)
    if values:
        values = list(values)
        for name, value in zip(names, values):
            try:
                settings[name] = ast.literal_eval(value)
            except (TypeError, ValueError, SyntaxError):
                settings[name] = value
        return True
    return False


def unload_modules(top_level_module_name):
    """Unloads all modules named starting with the specified string.

    This function eases the development workflow when editing a library that is
    used from Rhino/Grasshopper.

    Parameters
    ----------
    top_level_module_name : :obj:`str`
        Name of the top-level module to unload.

    Returns
    -------
    list
        List of unloaded module names.
    """
    modules = filter(lambda m: m.startswith(top_level_module_name), sys.modules)

    for module in modules:
        sys.modules.pop(module)

    return modules



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas_rhino
from compas.artists import SurfaceArtist
from compas.colors import Color
from .artist import RhinoArtist


class SurfaceArtist(RhinoArtist, SurfaceArtist):
    """Artist for drawing surfaces.

    Parameters
    ----------
    surface : :class:`compas.geometry.Surface`
        A COMPAS surface.
    layer : str, optional
        The layer that should contain the drawing.
    **kwargs : dict, optional
        Additional keyword arguments.
        For more info, see :class:`RhinoArtist` and :class:`~compas.artists.SurfaceArtist`.

    """

    def __init__(self, surface, layer=None, **kwargs):
        super(SurfaceArtist, self).__init__(surface=surface, layer=layer, **kwargs)

    def draw(self, color=None):
        """Draw the surface.

        Parameters
        ----------
        color : tuple[int, int, int] | tuple[float, float, float] | :class:`~compas.colors.Color`, optional
            The RGB color of the surface.
            The default color is :attr:`compas.artists.SurfaceArtist.color`.

        Returns
        -------
        list[System.Guid]
            The GUIDs of the created Rhino objects.

        """
        color = color or self.color
        if not Color.is_rgb255(color):
            color = Color(* list(color)).rgb255
        surfaces = [{'surface': self.surface, 'color': color, 'name': self.surface.name}]
        return compas_rhino.draw_surfaces(surfaces, layer=self.layer, clear=False, redraw=False)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

try:
    import Grasshopper as gh
except ImportError:
    pass


def update_component(ghenv, delay):
    """Schedule an update of the Grasshopper component.

    After the specified delay, the GH component will be automatically updated.

    Parameters
    ----------
    ghenv : :class:`GhPython.Component.PythonEnvironment`
        The current GHPython environment.
    delay : :obj:`int`
        Time in milliseconds until the update is performed.

    Raises
    ------
    ValueError
        If the delay is less than zero.

    """
    if delay <= 0:
        raise ValueError('Delay must be greater than zero')

    ghcomp = ghenv.Component
    ghdoc = ghcomp.OnPingDocument()

    def callback(ghdoc):
        if ghdoc.SolutionState != gh.Kernel.GH_ProcessStep.Process:
            ghcomp.ExpireSolution(False)

    ghdoc.ScheduleSolution(
        delay, gh.Kernel.GH_Document.GH_ScheduleDelegate(callback))



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from math import sqrt
from compas.geometry.primitives import Primitive
from compas.geometry.primitives import Vector
from compas.geometry.primitives import Point


class Plane(Primitive):
    """A plane is defined by a base point and a normal vector.

    Parameters
    ----------
    point : [float, float, float] | :class:`compas.geometry.Point`
        The base point of the plane.
    normal : [float, float, float] | :class:`compas.geometry.Vector`
        The normal vector of the plane.

    Attributes
    ----------
    point : :class:`compas.geometry.Plane`
        The base point of the plane.
    normal : :class:`compas.geometry.Vector`
        The normal vector of the plane.
    d : float, read-only
        The *d* parameter of the linear equation describing the plane.
    abcd : list[float], read-only
        The coefficients of the plane equation.

    Examples
    --------
    >>> plane = Plane([0, 0, 0], [0, 0, 1])
    >>> plane.point
    Point(0.000, 0.000, 0.000)
    >>> plane.normal
    Vector(0.000, 0.000, 1.000)

    """

    __slots__ = ['_point', '_normal']

    def __init__(self, point, normal, **kwargs):
        super(Plane, self).__init__(**kwargs)
        self._point = None
        self._normal = None
        self.point = point
        self.normal = normal

    # ==========================================================================
    # data
    # ==========================================================================

    @property
    def DATASCHEMA(self):
        """:class:`schema.Schema` : Schema of the data representation."""
        from schema import Schema
        return Schema({
            'point': Point.DATASCHEMA.fget(None),
            'normal': Vector.DATASCHEMA.fget(None)
        })

    @property
    def JSONSCHEMANAME(self):
        """str : Name of the schema of the data representation in JSON format."""
        return 'plane'

    @property
    def data(self):
        """dict : The data dictionary that represents the plane."""
        return {'point': self.point.data,
                'normal': self.normal.data}

    @data.setter
    def data(self, data):
        self.point = Point.from_data(data['point'])
        self.normal = Vector.from_data(data['normal'])

    @classmethod
    def from_data(cls, data):
        """Construct a plane from its data representation.

        Parameters
        ----------
        data : dict
            The data dictionary.

        Returns
        -------
        :class:`compas.geometry.Plane`
            The constructed plane.

        Examples
        --------
        >>> plane = Plane.from_data({'point': [0.0, 0.0, 0.0], 'normal': [0.0, 0.0, 1.0]})
        >>> plane.point
        Point(0.000, 0.000, 0.000)
        >>> plane.normal
        Vector(0.000, 0.000, 1.000)

        """
        return cls(Point.from_data(data['point']), Vector.from_data(data['normal']))

    # ==========================================================================
    # properties
    # ==========================================================================

    @property
    def point(self):
        return self._point

    @point.setter
    def point(self, point):
        self._point = Point(*point)

    @property
    def normal(self):
        return self._normal

    @normal.setter
    def normal(self, vector):
        self._normal = Vector(*vector)
        self._normal.unitize()

    @property
    def d(self):
        a, b, c = self.normal
        x, y, z = self.point
        return - a * x - b * y - c * z

    @property
    def abcd(self):
        a, b, c = self.normal
        d = self.d
        return a, b, c, d

    # ==========================================================================
    # customization
    # ==========================================================================

    def __repr__(self):
        return 'Plane({0!r}, {1!r})'.format(self.point, self.normal)

    def __len__(self):
        return 2

    def __getitem__(self, key):
        if key == 0:
            return self.point
        if key == 1:
            return self.normal
        raise KeyError

    def __setitem__(self, key, value):
        if key == 0:
            self.point = value
            return
        if key == 1:
            self.normal = value
            return
        raise KeyError

    def __iter__(self):
        return iter([self.point, self.normal])

    def __eq__(self, other):
        return self.point == other[0] and self.normal == other[1]

    # ==========================================================================
    # constructors
    # ==========================================================================

    @classmethod
    def from_three_points(cls, a, b, c):
        """Construct a plane from three points in three-dimensional space.

        Parameters
        ----------
        a : [float, float, float] | :class:`compas.geometry.Point`
            The first point.
        b : [float, float, float] | :class:`compas.geometry.Point`
            The second point.
        c : [float, float, float] | :class:`compas.geometry.Point`
            The second point.

        Returns
        -------
        :class:`compas.geometry.Plane`
            A plane with base point `a` and normal vector defined as the unitized
            cross product of the vectors `ab` and `ac`.

        Examples
        --------
        >>> plane = Plane.from_three_points([0.0, 0.0, 0.0], [2.0, 1.0, 0.0], [0.0, 3.0, 0.0])
        >>> plane.point
        Point(0.000, 0.000, 0.000)
        >>> plane.normal
        Vector(0.000, 0.000, 1.000)

        """
        a = Point(*a)
        b = Point(*b)
        c = Point(*c)
        normal = Vector.cross(b - a, c - a)
        return cls(a, normal)

    @classmethod
    def from_point_and_two_vectors(cls, point, u, v):
        """Construct a plane from a base point and two vectors.

        Parameters
        ----------
        point : [float, float, float] | :class:`compas.geometry.Point`
            The base point.
        u : [float, float, float] | :class:`compas.geometry.Vector`
            The first vector.
        v : [float, float, float] | :class:`compas.geometry.Vector`
            The second vector.

        Returns
        -------
        :class:`compas.geometry.Plane`
            A plane with base point `point` and normal vector defined as the unitized
            cross product of vectors `u` and `v`.

        Examples
        --------
        >>> plane = Plane.from_three_points([0.0, 0.0, 0.0], [2.0, 1.0, 0.0], [0.0, 3.0, 0.0])
        >>> plane.point
        Point(0.000, 0.000, 0.000)
        >>> plane.normal
        Vector(0.000, 0.000, 1.000)

        """
        normal = Vector.cross(u, v)
        return cls(point, normal)

    def from_abcd(cls, abcd):
        """Construct a plane from the plane equation coefficients.

        Parameters
        ----------
        abcd : [float, float, float, float]
            The equation coefficients.

        Returns
        -------
        :class:`compas.geometry.Plane`

        """
        a, b, c, d = abcd
        x = 1 / sqrt(a**2 + b**2 + c**2)
        normal = [a, b, c]
        point = [a * d * x, b * d * x, c * d * x]
        return cls(point, normal)

    @classmethod
    def worldXY(cls):
        """Construct the world XY plane.

        Returns
        -------
        :class:`compas.geometry.Plane`
            The world XY plane.

        """
        return cls([0, 0, 0], [0, 0, 1])

    @classmethod
    def from_frame(cls, frame):
        """Construct a plane from a frame.

        Returns
        -------
        :class:`compas.geometry.Plane`
            A plane with the frame's `point` and the frame's `normal`.

        Examples
        --------
        >>> from compas.geometry import Frame
        >>> frame = Frame([1, 1, 1], [0.68, 0.68, 0.27], [-0.67, 0.73, -0.15])
        >>> Plane.from_frame(frame)
        Plane(Point(1.000, 1.000, 1.000), Vector(-0.299, -0.079, 0.951))

        """
        return cls(frame.point, frame.normal)

    # ==========================================================================
    # methods
    # ==========================================================================

    def transform(self, T):
        """Transform this plane.

        Parameters
        ----------
        T : :class:`compas.geometry.Transformation` | list[list[float]]
            The transformation.

        Returns
        -------
        None

        Examples
        --------
        >>> from compas.geometry import Frame
        >>> from compas.geometry import Transformation
        >>> from compas.geometry import Plane
        >>> f = Frame([1, 1, 1], [0.68, 0.68, 0.27], [-0.67, 0.73, -0.15])
        >>> T = Transformation.from_frame(f)
        >>> plane = Plane.worldXY()
        >>> plane.transform(T)

        """
        self.point.transform(T)
        self.normal.transform(T)

    def offset(self, distance):
        """Returns a new offset plane by a given distance.
        Plane normal is used as positive direction.

        Parameters
        ----------
        distance: float
            The offset distance.

        Returns
        -------
        :class:`compas.geometry.Plane`
            The offset plane.

        """
        return Plane(self.point + self.normal.scaled(distance), self.normal)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import math

from compas.geometry import quaternion_multiply
from compas.geometry import quaternion_conjugate
from compas.geometry import quaternion_unitize
from compas.geometry import quaternion_canonize
from compas.geometry import quaternion_norm
from compas.geometry import quaternion_is_unit
from compas.geometry import quaternion_from_matrix

from compas.geometry.primitives import Primitive


class Quaternion(Primitive):
    r"""A quaternion is defined by 4 components, X, Y, Z, and W.

    Parameters
    ----------
    w : float
        The scalar (real) part of a quaternion.
    x, y, z : float
        Components of the vector (complex, imaginary) part of a quaternion.

    Attributes
    ----------
    w : float
        The W component of the quaternion.
    x : float
        The X component of the quaternion.
    y : float
        The Y component of the quaternion.
    z : float
        The Z component of the quaternion.
    wxyz : list[float], read-only
        Quaternion as a list of float in the 'wxyz' convention.
    xyzw : list[float], read-only
        Quaternion as a list of float in the 'xyzw' convention.
    norm : float, read-only
        The length (euclidean norm) of the quaternion.
    is_unit : bool, read-only
        True if the quaternion is unit-length.
        False otherwise.

    Notes
    -----
    The default convention to represent a quaternion :math:`q` in this module
    is by four real values :math:`w`, :math:`x`, :math:`y`, :math:`z`.
    The first value :math:`w` is the scalar (real) part,
    and :math:`x`, :math:`y`, :math:`z` form the vector (complex, imaginary) part [1]_, so that:

    .. math::

        q = w + xi + yj + zk

    where :math:`i, j, k` are basis components with following multiplication rules [2]_:

    .. math::

        \begin{align}
        ii &= jj = kk = ijk = -1 \\
        ij &= k, \quad ji = -k \\
        jk &= i, \quad kj = -i \\
        ki &= j, \quad ik = -j
        \end{align}

    Quaternions are associative but not commutative.

    **Quaternion as rotation.**

    A rotation through an angle :math:`\theta` around an axis
    defined by a euclidean unit vector :math:`u = u_{x}i + u_{y}j + u_{z}k`
    can be represented as a quaternion:

    .. math::

        q = cos(\frac{\theta}{2}) + sin(\frac{\theta}{2}) [u_{x}i + u_{y}j + u_{z}k]

    i.e.:

    .. math::

        \begin{align}
        w &= cos(\frac{\theta}{2}) \\
        x &= sin(\frac{\theta}{2}) u_{x} \\
        y &= sin(\frac{\theta}{2}) u_{y} \\
        z &= sin(\frac{\theta}{2}) u_{z}
        \end{align}

    For a quaternion to represent a rotation or orientation, it must be unit-length.
    A quaternion representing a rotation :math:`p` resulting from applying a rotation
    :math:`r` to a rotation :math:`q`, i.e.: :math:`p = rq`,
    is also unit-length.

    References
    ----------
    .. [1] http://mathworld.wolfram.com/Quaternion.html
    .. [2] http://mathworld.wolfram.com/HamiltonsRules.html
    .. [3] https://github.com/matthew-brett/transforms3d/blob/master/transforms3d/quaternions.py

    Examples
    --------
    >>> Q = Quaternion(1.0, 1.0, 1.0, 1.0).unitized()
    >>> R = Quaternion(0.0,-0.1, 0.2,-0.3).unitized()
    >>> P = R*Q
    >>> P.is_unit
    True

    """

    __slots__ = ['_w', '_x', '_y', '_z']

    def __init__(self, w, x, y, z, **kwargs):
        super(Quaternion, self).__init__(**kwargs)
        self._w = None
        self._x = None
        self._y = None
        self._z = None
        self.w = w
        self.x = x
        self.y = y
        self.z = z

    # ==========================================================================
    # data
    # ==========================================================================

    @property
    def DATASCHEMA(self):
        """:class:`schema.Schema` : Schema of the data representation."""
        from schema import Schema
        return Schema({'w': float, 'x': float, 'y': float, 'z': float})

    @property
    def JSONSCHEMANAME(self):
        """str : Name of the schema of the data representation in JSON format."""
        return 'quaternion'

    @property
    def data(self):
        """dict : Representation of the quaternion as a dict containing only native Python objects."""
        return {'w': self.w, 'x': self.x, 'y': self.y, 'z': self.z}

    @data.setter
    def data(self, data):
        self.w = data['w']
        self.x = data['x']
        self.y = data['y']
        self.z = data['z']

    @classmethod
    def from_data(cls, data):
        """Construct a quaternion from a data dict.

        Parameters
        ----------
        data : dict
            The data dictionary.

        Returns
        -------
        :class:`compas.geometry.Quaternion`
            The constructed quaternion.

        Examples
        --------
        >>>

        """
        return cls(data['w'], data['x'], data['y'], data['z'])

    # ==========================================================================
    # properties
    # ==========================================================================

    @property
    def w(self):
        return self._w

    @w.setter
    def w(self, w):
        self._w = float(w)

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        self._x = float(x)

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, y):
        self._y = float(y)

    @property
    def z(self):
        return self._z

    @z.setter
    def z(self, z):
        self._z = float(z)

    @property
    def wxyz(self):
        return [self.w, self.x, self.y, self.z]

    @property
    def xyzw(self):
        return [self.x, self.y, self.z, self.w]

    @property
    def norm(self):
        return quaternion_norm(self)

    @property
    def is_unit(self):
        return quaternion_is_unit(self)

    # ==========================================================================
    # customization
    # ==========================================================================

    def __getitem__(self, key):
        if key == 0:
            return self.w
        if key == 1:
            return self.x
        if key == 2:
            return self.y
        if key == 3:
            return self.z
        raise KeyError

    def __setitem__(self, key, value):
        if key == 0:
            self.w = value
            return
        if key == 1:
            self.x = value
            return
        if key == 2:
            self.y = value
        if key == 3:
            self.z = value
        raise KeyError

    def __eq__(self, other, tol=1e-05):
        if not hasattr(other, '__iter__') or not hasattr(other, '__len__') or len(self) != len(other):
            return False
        for v1, v2 in zip(self, other):
            if math.fabs(v1 - v2) > tol:
                return False
        return True

    def __iter__(self):
        return iter(self.wxyz)

    def __len__(self):
        return 4

    def __repr__(self):
        return 'Quaternion({:.{prec}f}, {:.{prec}f}, {:.{prec}f}, {:.{prec}f})'.format(self.w, self.x, self.y, self.z, prec=3)

    def __mul__(self, other):
        """Multiply operator for two quaternions.

        Parameters
        ----------
        other : [float, float, float, float] | :class:`compas.geometry.Quaternion`
            A Quaternion.

        Returns
        -------
        :class:`compas.geometry.Quaternion`
            The product :math:`P = R * Q` of this quaternion (R) multiplied by other quaternion (Q).

        Notes
        -----
        Multiplication of two quaternions :math:`R * Q` can be interpreted as applying rotation R to an orientation Q,
        provided that both R and Q are unit-length.
        The result is also unit-length.
        Multiplication of quaternions is not commutative!

        Examples
        --------
        >>> Q = Quaternion(1.0, 1.0, 1.0, 1.0).unitized()
        >>> R = Quaternion(0.0,-0.1, 0.2,-0.3).unitized()
        >>> P = R*Q
        >>> P.is_unit
        True

        """
        p = quaternion_multiply(list(self), list(other))
        return Quaternion(*p)

    # ==========================================================================
    # constructors
    # ==========================================================================

    @classmethod
    def from_frame(cls, frame):
        """Creates a quaternion object from a frame.

        Parameters
        ----------
        frame : :class:`compas.geometry.Frame`

        Returns
        -------
        :class:`compas.geometry.Quaternion`
            The new quaternion.

        Examples
        --------
        >>> from compas.geometry import allclose
        >>> from compas.geometry import Frame
        >>> q = [1., -2., 3., -4.]
        >>> F = Frame.from_quaternion(q)
        >>> Q = Quaternion.from_frame(F)
        >>> allclose(Q.canonized(), quaternion_canonize(quaternion_unitize(q)))
        True

        """
        w, x, y, z = frame.quaternion
        return cls(w, x, y, z)

    @classmethod
    def from_matrix(cls, M):
        """Create a Quaternion from a transformation matrix.

        Parameters
        ----------
        M : list[list[float]]

        Returns
        -------
        :class:`compas.geometry.Quaternion`
            The new quaternion.

        Example
        -------
        >>> from compas.geometry import matrix_from_euler_angles
        >>> ea = [0.2, 0.6, 0.2]
        >>> M = matrix_from_euler_angles(ea)
        >>> Quaternion.from_matrix(M)
        Quaternion(0.949, 0.066, 0.302, 0.066)

        """
        return cls(*quaternion_from_matrix(M))

    @classmethod
    def from_rotation(cls, R):
        """Create a Quaternion from a Rotatation.

        Parameters
        ----------
        R : :class:`compas.geometry.Rotation`

        Returns
        -------
        :class:`compas.geometry.Quaternion`
            The new quaternion.

        Example
        -------
        >>> from compas.geometry import Frame, Rotation
        >>> R = Rotation.from_frame(Frame.worldYZ())
        >>> Quaternion.from_rotation(R)
        Quaternion(0.500, 0.500, 0.500, 0.500)

        """
        return cls.from_matrix(R.matrix)

    # ==========================================================================
    # methods
    # ==========================================================================

    def unitize(self):
        """Scales the quaternion to make it unit-length.

        Returns
        -------
        None

        Examples
        --------
        >>> q = Quaternion(1.0, 1.0, 1.0, 1.0)
        >>> q.is_unit
        False
        >>> q.unitize()
        >>> q.is_unit
        True

        """
        qu = quaternion_unitize(self)
        self.w, self.x, self.y, self.z = qu

    def unitized(self):
        """Returns a quaternion with a unit-length.

        Returns
        -------
        :class:`compas.geometry.Quaternion`

        Examples
        --------
        >>> q = Quaternion(1.0, 1.0, 1.0, 1.0)
        >>> q.is_unit
        False
        >>> p = q.unitized()
        >>> p.is_unit
        True

        """
        qu = quaternion_unitize(self)
        return Quaternion(*qu)

    def canonize(self):
        """Makes the quaternion canonic.

        Returns
        -------
        None

        Examples
        --------
        >>> from compas.geometry import Frame
        >>> q = Quaternion.from_frame(Frame.worldZX())
        >>> q
        Quaternion(-0.500, 0.500, 0.500, 0.500)
        >>> q.canonize()
        >>> q
        Quaternion(0.500, -0.500, -0.500, -0.500)

        """
        qc = quaternion_canonize(self)
        self.w, self.x, self.y, self.z = qc

    def canonized(self):
        """Returns a quaternion in canonic form.

        Returns
        -------
        :class:`compas.geometry.Quaternion`
            A quaternion in canonic form.

        Examples
        --------
        >>> from compas.geometry import Frame
        >>> q = Quaternion.from_frame(Frame.worldZX())
        >>> q
        Quaternion(-0.500, 0.500, 0.500, 0.500)
        >>> p = q.canonized()
        >>> p
        Quaternion(0.500, -0.500, -0.500, -0.500)

        """
        qc = quaternion_canonize(self)
        return Quaternion(*qc)

    def conjugate(self):
        """Conjugate the quaternion.

        Returns
        -------
        None

        Examples
        --------
        >>> q = Quaternion(1.0, 1.0, 1.0, 1.0)
        >>> q.conjugate()
        >>> q
        Quaternion(1.000, -1.000, -1.000, -1.000)

        """
        qc = quaternion_conjugate(self)
        self.w, self.x, self.y, self.z = qc

    def conjugated(self):
        """Returns a conjugate quaternion.

        Returns
        -------
        :class:`compas.geometry.Quaternion`
            The conjugated quaternion.

        Examples
        --------
        >>> q = Quaternion(1.0, 1.0, 1.0, 1.0)
        >>> p = q.conjugated()
        >>> q
        Quaternion(1.000, 1.000, 1.000, 1.000)
        >>> p
        Quaternion(1.000, -1.000, -1.000, -1.000)

        """
        qc = quaternion_conjugate(self)
        return Quaternion(*qc)



import sys


__all__ = [
    'unload_modules',
]


def unload_modules(top_level_module_name):
    """Unloads all modules named starting with the specified string.

    This function eases the development workflow when editing a library that is
    used from Blender.

    Parameters
    ----------
    top_level_module_name : str
        Name of the top-level module to unload.

    Returns
    -------
    list[str]
        List of unloaded module names.

    """
    modules = list(filter(lambda m: m.startswith(top_level_module_name), sys.modules))

    for module in modules:
        sys.modules.pop(module)

    return modules



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

try:
    basestring
except NameError:
    basestring = str

import colorsys
import re

from compas.colors.html_colors import HTML_TO_RGB255
from compas.data import Data

BASE16 = '0123456789abcdef'

try:
    HEX_DEC = {v: int(v, base=16) for v in [x + y for x in BASE16 for y in BASE16]}
except Exception:
    HEX_DEC = {v: int(v, 16) for v in [x + y for x in BASE16 for y in BASE16]}


class Color(Data):
    """Class for working with colors.

    Parameters
    ----------
    red : float
        The red component in the range of 0-1.
    green : float
        The green component in the range of 0-1.
    blue : float
        The blue component in the range of 0-1.
    alpha : float, optional
        Transparency setting.
        If ``alpha = 0.0``, the color is fully transparent.
        If ``alpha = 1.0``, the color is fully opaque.

    Other Parameters
    ----------------
    **kwargs : dict, optional
        See :class:`Data` for more information.

    Attributes
    ----------
    r : float
        Red component of the color in RGB1 color space.
    g : float
        Green component of the color in RGB1 color space.
    b : float
        Blue component of the color in RGB1 color space.
    rgb : tuple[float, float, float]
        RGB1 color tuple, with components in the range 0-1.
    rgb255 : tuple[int, int, int]
        RGB255 color tuple, with components in the range 0-255.
    hex : str
        Hexadecimal color string.
    hls : tuple[float, float, float]
        Hue, Lightness, Saturation.
    hsv : tuple[float, float, float]
        Hue, Saturation, Value / Brightness.
    lightness : float
        How much white the color appears to contain.
        This is the "Lightness" in HLS.
        Making a color "lighter" is like adding more white.
    brightness : float
        How well-lit the color appears to be.
        This is the "Value" in HSV.
        Making a color "brighter" is like shining a stronger light on it, or illuminating it better.
    is_light : bool
        If True, the color is considered light.

    Examples
    --------
    >>> Color(1, 0, 0)
    Color(1.0, 0.0, 0.0, 1.0)
    >>> Color.red()
    Color(1.0, 0.0, 0.0, 1.0)
    >>> Color(1, 0, 0) == Color.red()
    True

    >>> Color.magenta()
    Color(1.0, 0.0, 1.0, 1.0)
    >>> Color.lime()
    Color(0.5, 1.0, 0.0, 1.0)
    >>> Color.navy()
    Color(0.0, 0.0, 0.5, 1.0)
    >>> Color.olive()
    Color(0.5, 0.5, 0.0, 1.0)

    >>> Color.lime().is_light
    True
    >>> Color.navy().is_light
    False

    """

    def __init__(self, red, green, blue, alpha=1.0, **kwargs):
        super(Color, self).__init__(**kwargs)
        self._r = 1.0
        self._g = 1.0
        self._b = 1.0
        self._a = 1.0
        self.r = red
        self.g = green
        self.b = blue
        self.a = alpha

    # --------------------------------------------------------------------------
    # data
    # --------------------------------------------------------------------------

    @property
    def data(self):
        return {'red': self.r, 'green': self.g, 'blue': self.b, 'alpha': self.a}

    @data.setter
    def data(self, data):
        self.r = data['red']
        self.g = data['green']
        self.b = data['blue']
        self.a = data['alpha']

    @classmethod
    def from_data(cls, data):
        return cls(data['red'], data['green'], data['blue'], data['alpha'])

    # --------------------------------------------------------------------------
    # properties
    # --------------------------------------------------------------------------

    @property
    def r(self):
        return self._r

    @r.setter
    def r(self, red):
        if red > 1.0 or red < 0.0:
            raise ValueError("Components of an RGBA color should be in the range 0-1.")
        self._r = float(red)

    @property
    def g(self):
        return self._g

    @g.setter
    def g(self, green):
        if green > 1.0 or green < 0.0:
            raise ValueError("Components of an RGBA color should be in the range 0-1.")
        self._g = float(green)

    @property
    def b(self):
        return self._b

    @b.setter
    def b(self, blue):
        if blue > 1.0 or blue < 0.0:
            raise ValueError("Components of an RGBA color should be in the range 0-1.")
        self._b = float(blue)

    @property
    def a(self):
        return self._a

    @a.setter
    def a(self, alpha):
        if alpha > 1.0 or alpha < 0.0:
            raise ValueError("Components of an RGBA color should be in the range 0-1.")
        self._a = float(alpha)

    @property
    def rgb(self):
        r = self.r
        g = self.g
        b = self.b
        return r, g, b

    @property
    def rgb255(self):
        r = int(self.r * 255)
        g = int(self.g * 255)
        b = int(self.b * 255)
        return r, g, b

    @property
    def rgba(self):
        r, g, b = self.rgb
        a = self.a
        return r, g, b, a

    @property
    def rgba255(self):
        r, g, b = self.rgb255
        a = int(self.a * 255)
        return r, g, b, a

    @property
    def hex(self):
        return '#{0:02x}{1:02x}{2:02x}'.format(* self.rgb255)

    @property
    def hls(self):
        return colorsys.rgb_to_hls(* self.rgb)

    @property
    def hsv(self):
        return colorsys.rgb_to_hsv(* self.rgb)

    @property
    def lightness(self):
        return self.hls[1]

    @property
    def brightness(self):
        return self.hsv[2]

    @property
    def is_light(self):
        return self.luminance > 0.179

    @property
    def yuv(self):
        """tuple[float, float, float] :
        Luma and chroma components, with chroma defined by the blue and red projections.
        """
        y = self.luma
        u, v = self.chroma
        return y, u, v

    @property
    def luma(self):
        """float :
        The brightness of a yuv signal.
        """
        return 0.299 * self.r + 0.587 * self.g + 0.114 * self.b

    @property
    def chroma(self):
        """tuple[float, float] :
        The color of a yuv signal.
        "How different from a grey of the same lightness the color appears to be."
        """
        y = self.luma
        u = 0.492 * (self.b - y)
        v = 0.877 * (self.r - y)
        return u, v

    @property
    def luminance(self):
        """float :
        The amount of light that passes through, is emitted from, or is reflected from a particular area.
        Here, it expresses the preceived brightness of the color.
        Note that this is not the same as the "Lightness" of HLS or the "Value/Brightness" of HSV.
        """
        return 0.2126 * self.r + 0.7152 * self.g + 0.0722 * self.b

    @property
    def saturation(self):
        """float : The perceived freedom of whiteness."""
        maxval = max(self.r, self.g, self.b)
        minval = min(self.r, self.g, self.b)
        return (maxval - minval) / maxval

    # --------------------------------------------------------------------------
    # customization
    # --------------------------------------------------------------------------

    def __repr__(self):
        return 'Color({}, {}, {}, {})'.format(self.r, self.g, self.b, self.a)

    def __getitem__(self, key):
        if key == 0:
            return self.r
        if key == 1:
            return self.g
        if key == 2:
            return self.b
        raise KeyError

    def __len__(self):
        return 3

    def __iter__(self):
        return iter(self.rgb)

    def __eq__(self, other):
        return all(a == b for a, b in zip(self, other))

    # --------------------------------------------------------------------------
    # constructors
    # --------------------------------------------------------------------------

    @classmethod
    def from_rgb255(cls, r, g, b):
        """Construct a color from RGB255 components.

        Parameters
        ----------
        r : int & valuerange[0, 255]
            Red component.
        g : int & valuerange[0, 255]
            Green component.
        b : int & valuerange[0, 255]
            Blue component.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(r/255, g/255, b/255)

    @classmethod
    def from_hls(cls, h, l, s):  # noqa: E741
        """Construct a color from Hue, Luminance, and Saturation.

        Parameters
        ----------
        h : float
            Hue.
        l : float
            Luminance.
        s : float
            Saturation.

        Returns
        -------
        :class:`compas.colors.Color`

        See Also
        --------
        https://en.wikipedia.org/wiki/HSL_and_HSV

        """
        r, g, b = colorsys.hls_to_rgb(h, l, s)
        return cls(r, g, b)

    @classmethod
    def from_hsv(cls, h, s, v):
        """Construct a color from Hue, Saturation, and Value.

        Parameters
        ----------
        h : float
            Hue.
        s : float
            Saturation.
        v : float
            Value.

        Returns
        -------
        :class:`compas.colors.Color`

        See Also
        --------
        https://en.wikipedia.org/wiki/HSL_and_HSV

        """
        r, g, b = colorsys.hsv_to_rgb(h, s, v)
        return cls(r, g, b)

    @classmethod
    def from_yiq(cls, y, i, q):
        """Construct a color from components in the YIQ color space.

        Parameters
        ----------
        y : float
            Luma.
        i : float
            Orange-blue chroma.
        q : float
            Purple-green chroma.

        Returns
        -------
        :class:`compas.colors.Color`

        See Also
        --------
        https://en.wikipedia.org/wiki/YIQ

        """
        r, g, b = colorsys.yiq_to_rgb(y, i, q)
        return cls(r, g, b)

    @classmethod
    def from_yuv(cls, y, u, v):
        """Construct a color from components in the YUV color space.

        Parameters
        ----------
        y : float
            Luma.
        u : float
            Blue projection chroma.
        v : float
            Red projection chroma.

        Returns
        -------
        :class:`compas.colors.Color`

        See Also
        --------
        https://en.wikipedia.org/wiki/YUV

        """
        r = y + 1.140 * v
        g = y - 0.395 * u - 0.581 * v
        b = y + 2.032 * u
        return cls(r, g, b)

    @classmethod
    def from_i(cls, i):
        """Construct a color from a single number in the range 0-1.

        Parameters
        ----------
        i : float
            Number in the range 0-1, representing the color.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        if i == 0.0:
            r, g, b = 0, 0, 255
        elif 0.0 < i < 0.25:
            r, g, b = 0, int(255 * (4 * i)), 255
        elif i == 0.25:
            r, g, b = 0, 255, 255
        elif 0.25 < i < 0.5:
            r, g, b = 0, 255, int(255 - 255 * 4 * (i - 0.25))
        elif i == 0.5:
            r, g, b = 0, 255, 0
        elif 0.5 < i < 0.75:
            r, g, b = int(0 + 255 * 4 * (i - 0.5)), 255, 0
        elif i == 0.75:
            r, g, b = 255, 255, 0
        elif 0.75 < i < 1.0:
            r, g, b,  = 255, int(255 - 255 * 4 * (i - 0.75)), 0
        elif i == 1.0:
            r, g, b = 255, 0, 0
        else:
            r, g, b = 0, 0, 0
        return cls(r / 255.0, g / 255.0, b / 255.0)

    @classmethod
    def from_hex(cls, value):
        """Construct a color from a hexadecimal color value.

        Parameters
        ----------
        value : str
            The hexadecimal color.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        value = value.lstrip('#').lower()
        r = HEX_DEC[value[0:2]]
        g = HEX_DEC[value[2:4]]
        b = HEX_DEC[value[4:6]]
        return cls(r / 255.0, g / 255.0, b / 255.0)

    @classmethod
    def from_name(cls, name):
        """Construct a color from a name in the extended color table of HTML/CSS/SVG.

        Parameters
        ----------
        name : str
            The color name. The name is case-insensitive.

        Returns
        -------
        :class:`compas.colors.Color`

        See Also
        --------
        https://www.w3.org/TR/css-color-3/#svg-color
        """
        rgb255 = HTML_TO_RGB255.get(name.lower())
        if rgb255 is None:
            raise ValueError("Color name not found.")
        return cls.from_rgb255(*rgb255)

    # --------------------------------------------------------------------------
    # presets
    # --------------------------------------------------------------------------

    @classmethod
    def white(cls):
        """Construct the color white.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(1.0, 1.0, 1.0)

    @classmethod
    def black(cls):
        """Construct the color black.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.0, 0.0, 0.0)

    @classmethod
    def grey(cls):
        """Construct the color grey.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.5, 0.5, 0.5)

    @classmethod
    def red(cls):
        """Construct the color red.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(1.0, 0.0, 0.0)

    @classmethod
    def orange(cls):
        """Construct the color orange.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(1.0, 0.5, 0.0)

    @classmethod
    def yellow(cls):
        """Construct the color yellow.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(1.0, 1.0, 0.0)

    @classmethod
    def lime(cls):
        """Construct the color lime (or chartreuse green).

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.5, 1.0, 0.0)

    @classmethod
    def green(cls):
        """Construct the color green.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.0, 1.0, 0.0)

    @classmethod
    def mint(cls):
        """Construct the color mint (or spring green).

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.0, 1.0, 0.5)

    @classmethod
    def cyan(cls):
        """Construct the color cyan.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.0, 1.0, 1.0)

    @classmethod
    def azure(cls):
        """Construct the color azure.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.0, 0.5, 1.0)

    @classmethod
    def blue(cls):
        """Construct the color blue.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.0, 0.0, 1.0)

    @classmethod
    def violet(cls):
        """Construct the color violet.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.5, 0.0, 1.0)

    @classmethod
    def magenta(cls):
        """Construct the color magenta.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(1.0, 0.0, 1.0)

    @classmethod
    def pink(cls):
        """Construct the color pink.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(1.0, 0.0, 0.5)

    # --------------------------------------------------------------------------
    # other presets
    # --------------------------------------------------------------------------

    @classmethod
    def maroon(cls):
        """Construct the color maroon.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.5, 0.0, 0.0)

    @classmethod
    def brown(cls):
        """Construct the color brown.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.5, 0.25, 0.0)

    @classmethod
    def olive(cls):
        """Construct the color olive.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.5, 0.5, 0.0)

    @classmethod
    def teal(cls):
        """Construct the color teal.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.0, 0.5, 0.5)

    @classmethod
    def navy(cls):
        """Construct the color navy.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.0, 0.0, 0.5)

    @classmethod
    def purple(cls):
        """Construct the color purple.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.5, 0.0, 0.5)

    @classmethod
    def silver(cls):
        """Construct the color silver.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        return cls(0.75, 0.75, 0.75)

    # ochre
    # beige
    # bordeaux
    # hotpink
    # steel
    # midnight
    # salmon

    # --------------------------------------------------------------------------
    # methods
    # --------------------------------------------------------------------------

    def is_rgb1(self):
        """Verify that the color is in the RGB 1 color space.

        Returns
        -------
        bool

        """
        return all(isinstance(c, float) and (c >= 0 and c <= 1) for c in self)

    def is_rgb255(self):
        """Verify that the color is in the RGB 255 color space.

        Returns
        -------
        bool

        """
        return all(isinstance(c, int) and (c >= 0 and c <= 255) for c in self)

    def is_hex(self):
        """Verify that the color is in hexadecimal format.

        Returns
        -------
        bool

        """
        if isinstance(self, basestring):
            match = re.search(r'^#(?:[0-9a-fA-F]{3}){1,2}$', self)
            if match:
                return True
            return False
        return False

    def lighten(self, factor=10):
        """Lighten the color.

        Parameters
        ----------
        factor : float, optional
            Percentage of lightness increase.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the percentage of lightness increase is not in the range 0-100.

        """
        if factor > 100 or factor < 0:
            raise ValueError('Percentage of increased lightness should be in the range 0-100.')

        factor = 1.0 + factor / 100

        h, l, s = self.hls
        r, g, b = colorsys.hls_to_rgb(h, min(1.0, l * factor), s)
        self.r = r
        self.g = g
        self.b = b

    def lightened(self, factor=10):
        """Return a lightened copy of the color.

        Parameters
        ----------
        factor : float, optional
            Percentage of lightness increase.

        Returns
        -------
        :class:`compas.colors.Color`

        Raises
        ------
        ValueError
            If the percentage of lightness increase is not in the range 0-100.

        """
        color = self.copy()
        color.lighten(factor=factor)
        return color

    def darken(self, factor=10):
        """Darken the color.

        Parameters
        ----------
        factor : float, optional
            Percentage of lightness reduction.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the percentage of lightness reduction is not in the range 0-100.

        """
        if factor > 100 or factor < 0:
            raise ValueError('Percentage of reduced lightness should be in the range 0-100.')

        factor = 1.0 - factor / 100

        h, l, s = self.hls
        r, g, b = colorsys.hls_to_rgb(h, max(0.0, l * factor), s)
        self.r = r
        self.g = g
        self.b = b

    def darkened(self, factor=10):
        """Return a darkened copy of the color.

        Parameters
        ----------
        factor : float, optional
            Percentage of lightness reduction.

        Returns
        -------
        :class:`compas.colors.Color`

        Raises
        ------
        ValueError
            If the percentage of lightness reduction is not in the range 0-100.

        """
        color = self.copy()
        color.darken(factor=factor)
        return color

    def invert(self):
        """Invert the current color wrt to the RGB color circle.

        Returns
        -------
        None

        """
        self.r = 1.0 - self.r
        self.g = 1.0 - self.g
        self.b = 1.0 - self.b

    def inverted(self):
        """Return an inverted copy of the color.

        Returns
        -------
        :class:`compas.colors.Color`

        """
        color = self.copy()
        color.invert()
        return color

    def saturate(self, factor=10):
        """Saturate the color by a given percentage.

        Parameters
        ----------
        factor : float, optional
            Percentage of saturation increase.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the percentage of saturation is not in the range 0-100.

        """
        if factor > 100 or factor < 0:
            raise ValueError('Percentage of saturation should be in the range 0-100.')

        factor = 1.0 + factor / 100

        h, l, s = self.hls
        r, g, b = colorsys.hls_to_rgb(h, l, min(1.0, s * factor))
        self.r = r
        self.g = g
        self.b = b

    def saturated(self, factor=10):
        """Return a saturated copy of the color.

        Parameters
        ----------
        factor : float, optional
            Percentage of saturation increase.

        Returns
        -------
        :class:`compas.colors.Color`

        Raises
        ------
        ValueError
            If the percentage of desaturation is not in the range 0-100.

        """
        color = self.copy()
        color.saturate(factor=factor)
        return color

    def desaturate(self, factor=10):
        """Desaturate the color by a given percentage.

        Parameters
        ----------
        factor : float, optional
            Percentage of saturation reduction.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the percentage of desaturation is not in the range 0-100.

        """
        if factor > 100 or factor < 0:
            raise ValueError('Percentage of desaturation should be in the range 0-100.')

        factor = 1.0 - factor / 100

        h, l, s = self.hls
        r, g, b = colorsys.hls_to_rgb(h, l, max(0.0, s * factor))
        self.r = r
        self.g = g
        self.b = b

    def desaturated(self, factor=10):
        """Return a desaturated copy of the color.

        Parameters
        ----------
        factor : float, optional
            Percentage of saturation reduction.

        Returns
        -------
        :class:`compas.colors.Color`

        Raises
        ------
        ValueError
            If the percentage of desaturation is not in the range 0-100.

        """
        color = self.copy()
        color.desaturate(factor=factor)
        return color



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import centroid_points


__all__ = [
    'network_smooth_centroid',
]


def network_smooth_centroid(network, fixed=None, kmax=100, damping=0.5, callback=None, callback_args=None):
    """Smooth a network by moving every free node to the centroid of its neighbors.

    Parameters
    ----------
    network : Mesh
        A network object.
    fixed : list, optional
        The fixed nodes of the network.
    kmax : int, optional
        The maximum number of iterations.
    damping : float, optional
        The damping factor.
    callback : callable, optional
        A user-defined callback function to be executed after every iteration.
    callback_args : list, optional
        A list of arguments to be passed to the callback.

    Returns
    -------
    None

    Raises
    ------
    Exception
        If a callback is provided, but it is not callable.

    """
    if callback:
        if not callable(callback):
            raise Exception('Callback is not callable.')

    fixed = fixed or []
    fixed = set(fixed)

    for k in range(kmax):
        key_xyz = {key: network.node_coordinates(key) for key in network.nodes()}

        for key, attr in network.nodes(True):
            if key in fixed:
                continue

            x, y, z = key_xyz[key]

            cx, cy, cz = centroid_points([key_xyz[nbr] for nbr in network.neighbors(key)])

            attr['x'] += damping * (cx - x)
            attr['y'] += damping * (cy - y)
            attr['z'] += damping * (cz - z)

        if callback:
            callback(k, callback_args)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.utilities import is_color_rgb
from .artist import Artist


class PrimitiveArtist(Artist):
    """Base class for artists for geometry primitives.

    Parameters
    ----------
    primitive: :class:`compas.geometry.Primitive`
        The geometry of the primitive.
    color : tuple[float, float, float], optional
        The RGB components of the base color of the primitive.

    Attributes
    ----------
    primitive : :class:`compas.geometry.Primitive`
        The geometric primitive associated with the artist.
    color : tuple[float, float, float]
        The color of the object.

    Class Attributes
    ----------------
    default_color : tuple[float, float, float]
        The default rgb color value of the primitive.

    """

    default_color = (0, 0, 0)

    def __init__(self, primitive, color=None, **kwargs):
        super(PrimitiveArtist, self).__init__()
        self._primitive = None
        self._color = None
        self.primitive = primitive
        self.color = color

    @property
    def primitive(self):
        return self._primitive

    @primitive.setter
    def primitive(self, primitive):
        self._primitive = primitive

    @property
    def color(self):
        if not self._color:
            self._color = self.default_color
        return self._color

    @color.setter
    def color(self, color):
        if is_color_rgb(color):
            self._color = color



"""
********************************************************************************
data
********************************************************************************

.. currentmodule:: compas.data

.. rst-class:: lead

This package provides a base data class for all COMPAS data objects such as geometry objects, robots, and data structures,
and the infrastructure for data validation, conversion, coercion, and JSON serialisation.


Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Data
    DataEncoder
    DataDecoder


Functions
=========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    json_load
    json_loads
    json_dump
    json_dumps


Validators
==========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    is_sequence_of_int
    is_sequence_of_uint
    is_sequence_of_float
    is_int3
    is_float3
    is_float4x4
    validate_data


Exceptions
==========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    DecoderError

"""
from __future__ import absolute_import

from .exceptions import DecoderError
from .validators import is_sequence_of_int
from .validators import is_sequence_of_uint
from .validators import is_sequence_of_float
from .validators import is_int3
from .validators import is_float3
from .validators import is_float4x4
from .validators import validate_data
from .encoders import DataEncoder
from .encoders import DataDecoder
from .data import Data

from .json import json_load, json_loads, json_dump, json_dumps

__all__ = [
    'Data',
    'DataEncoder',
    'DataDecoder',
    'DecoderError',
    'is_sequence_of_int',
    'is_sequence_of_uint',
    'is_sequence_of_float',
    'is_int3',
    'is_float3',
    'is_float4x4',
    'json_load',
    'json_loads',
    'json_dump',
    'json_dumps',
    'validate_data'
]
